from .mock_storage import sessions


class SessionStorageMock:

    async def load(self, session_id):
        for item in sessions:
            if session_id in item:
                return item[session_id]
        return None

    async def create(self, data):
        for item in data:
            sessions.append({item['id']: item})



from typing import List

import elasticsearch

from tracardi.domain.storage_result import StorageResult
from tracardi.domain.value_object.bulk_insert_result import BulkInsertResult
from tracardi.service.storage import index
from tracardi.service.storage.elastic import Elastic
from tracardi.service.storage.index import Index


class ElasticStorage:

    def __init__(self, index_key):
        self.storage = Elastic.instance()
        self.index = index.resources[index_key]  # type: Index

    async def load(self, id) -> [dict, None]:

        try:
            index = self.index.get_read_index()
            if not self.index.multi_index:
                result = await self.storage.get(index, id)
                output = result['_source']
                output['id'] = result['_id']
            else:
                query = {
                    "query": {
                        "term": {
                            "_id": id
                        }
                    }
                }
                result = StorageResult(await self.storage.search(index, query))
                output = list(result)
                if len(output) != 1:
                    return None
                output = output[0]

            return output
        except elasticsearch.exceptions.NotFoundError:
            return None

    async def create(self, payload) -> BulkInsertResult:
        return await self.storage.insert(self.index.get_write_index(), payload)

    async def delete(self, id):
        if not self.index.multi_index:
            return await self.storage.delete(self.index.get_read_index(), id)
        else:
            return await self.delete_by("_id", id)

    async def search(self, query):
        return await self.storage.search(self.index.get_read_index(), query)

    async def translate(self, sql):
        return await self.storage.translate(sql)

    async def refresh(self, params=None, headers=None):
        return await self.storage.refresh(self.index.get_write_index(), params, headers)

    async def load_by(self, field, value):
        query = {
            "query": {
                "term": {
                    field: value
                }
            }
        }
        return await self.search(query)

    async def delete_by(self, field, value):
        query = {
            "query": {
                "term": {
                    field: value
                }
            }
        }
        return await self.storage.delete_by_query(self.index.get_read_index(), query)

    async def load_by_values(self, fields_and_values: List[tuple], limit=1000):

        terms = []
        for field, value in fields_and_values:
            terms.append({
                "term": {
                    f"{field}": value
                }
            })

        query = {
            "size": limit,
            "query": {
                "bool": {
                    "must": terms
                }
            }
        }

        return await self.search(query)

    async def flush(self, params, headers):
        return await self.storage.flush(self.index.get_write_index(), params, headers)



from datetime import datetime
from typing import Optional, Any
from .entity import Entity
from tracardi.service.storage.crud import StorageCrud
from .metadata import Metadata
from .time import Time
from ..service.secrets import encrypt, decrypt


class Source(Entity):
    type: str
    metadata: Optional[Metadata]
    name: Optional[str] = "No name provided"
    description: Optional[str] = "No description provided"
    config: Optional[dict] = {}
    origin: str = "event"
    enabled: Optional[bool] = True
    consent: bool = False

    def __init__(self, **data: Any):
        data['metadata'] = Metadata(
            time=Time(
                insert=datetime.utcnow()
            ))
        super().__init__(**data)

    # Persistence
    def storage(self) -> StorageCrud:
        return StorageCrud("source", Source, entity=self)


class SourceRecord(Entity):
    type: str
    metadata: Optional[Metadata]
    name: Optional[str] = "No name provided"
    description: Optional[str] = "No description provided"
    config: Optional[str] = None
    enabled: Optional[bool] = True
    origin: str = "event"
    consent: bool = False

    def __init__(self, **data: Any):
        data['metadata'] = Metadata(
            time=Time(
                insert=datetime.utcnow()
            ))
        super().__init__(**data)

    @staticmethod
    def encode(source: Source) -> 'SourceRecord':
        return SourceRecord(
            id=source.id,
            name=source.name,
            description=source.description,
            type=source.type,
            origin=source.origin,
            enabled=source.enabled,
            consent=source.consent,
            config=encrypt(source.config)
        )

    def decode(self) -> Source:
        return Source(
            id=self.id,
            name=self.name,
            description=self.description,
            type=self.type,
            origin=self.origin,
            enabled=self.enabled,
            consent=self.consent,
            config=decrypt(self.config)
        )

    # Persistence
    def storage(self) -> StorageCrud:
        return StorageCrud("source", Source, entity=self)



from tracardi.domain.value_object.bulk_insert_result import BulkInsertResult


class SaveResult(BulkInsertResult):
    types: list = []

    def __add__(self, other):
        if not isinstance(other, SaveResult) and not isinstance(other, BulkInsertResult):
            raise ValueError("SaveResult can add only other SaveResult")

        self.saved += other.saved
        self.errors += other.errors
        self.ids += other.ids
        if isinstance(other, SaveResult):
            self.types += other.types

        return self

    def to_json(self):
        return {
            "saved": self.saved,
            "errors": self.errors,
            "ids": self.ids,
            "types": self.types
        }



from tracardi_dot_notation.dot_accessor import DotAccessor

from tracardi.process_engine.tql.parser import Parser
from tracardi.process_engine.tql.transformer.expr_transformer import ExprTransformer


class Condition:

    @staticmethod
    def parse(condition):
        parser = Parser(Parser.read('grammar/uql_expr.lark'), start='expr')
        return parser.parse(condition)

    @staticmethod
    def evaluate(condition, dot: DotAccessor):
        tree = Condition.parse(condition)
        return ExprTransformer(dot=dot).transform(tree)




from tracardi_plugin_sdk.domain.register import Plugin, Spec, MetaData
from tracardi_plugin_sdk.domain.result import Result
from tracardi_plugin_sdk.action_runner import ActionRunner

from tracardi.domain.event import Event
from tracardi.domain.profile import Profile
from tracardi.domain.session import Session
from tracardi_dot_notation.dot_accessor import DotAccessor


class AppendTraitAction(ActionRunner):

    def __init__(self, **kwargs):
        if 'append' not in kwargs:
            raise ValueError("Please define `append` in config section.")
        if not isinstance(kwargs['append'], dict):
            raise ValueError("Please define `append` as dictionary not {}.".format(type(kwargs['append'])))

        self.mapping = kwargs['append']

    async def run(self, payload: dict):

        dot = DotAccessor(
            self.profile,
            self.session,
            payload if isinstance(payload, dict) else None,
            self.event,
            self.flow)

        for destination, value in self.mapping.items():
            if destination in dot:
                if not isinstance(dot[destination], list):
                    dot[destination] = [dot[destination]]

                if value not in dot[destination]:
                    dot[destination].append(value)
            else:
                dot[destination] = value

        if not isinstance(dot.profile['traits']['private'], dict):
            raise ValueError("Error when appending profile@traits.private to value `{}`. Private must have key:value pair. "
                             "E.g. `name`: `{}`".format(dot.profile['traits']['private'], dot.profile['traits']['private']))

        if not isinstance(dot.profile['traits']['public'], dict):
            raise ValueError("Error when appending profile@traits.public to value `{}`. Public must have key:value pair. "
                             "E.g. `name`: `{}`".format(dot.profile['traits']['public'], dot.profile['traits']['public']))

        profile = Profile(**dot.profile)
        event = Event(**dot.event)
        session = Session(**dot.session)

        self.profile.replace(profile)
        self.session.replace(session)
        self.event.replace(event)

        return Result(port="payload", value=payload)


def register() -> Plugin:
    return Plugin(
        start=False,
        spec=Spec(
            module='tracardi.process_engine.action.v1.traits.append_trait_action',
            className='AppendTraitAction',
            inputs=['payload'],
            outputs=["payload"],
            init={
                "append": {
                    "target1": "source1",
                    "target2": "source2",
                }
            },
            version='0.1',
            license="MIT",
            author="Risto Kowaczewski"
        ),
        metadata=MetaData(
            name='Append Trait',
            desc='Appends trait if it a value already exists or sets trait if it does not.',
            type='flowNode',
            width=100,
            height=100,
            icon='append',
            group=["Traits"]
        )
    )



class Field:
    def __init__(self, label, value_ref):
        self.label = label
        self.dot = value_ref
        self.value = None

    def _get_value(self):
        if self.value is None:
            self.value = self.dot[self.label]
        return self.value

    def __eq__(self, other):
        return other == self._get_value()

    def __gt__(self, other):
        return self._get_value() > other

    def __ge__(self, other):
        return self._get_value() >= other

    def __lt__(self, other):
        return self._get_value() < other

    def __le__(self, other):
        return self._get_value() <= other



import asyncio

from tracardi_plugin_sdk.domain.register import Plugin, Spec, MetaData
from tracardi_plugin_sdk.domain.result import Result
from tracardi_plugin_sdk.action_runner import ActionRunner

from tracardi.domain.entity import Entity
from tracardi.domain.event import Event
from tracardi.domain.profile import Profile
from tracardi.domain.session import Session


class DebugPayloadAction(ActionRunner):

    def __init__(self, **kwargs):
        if 'event' not in kwargs or 'id' not in kwargs['event']:
            raise ValueError("Please define event.id in config section.")
        self.event_id = kwargs['event']['id']

    async def run(self, **kwargs):
        if self.debug:

            event_entity = Entity(id=self.event_id)
            event = await event_entity.storage('event').load(Event)  # type: Event

            if event is None:
                raise ValueError(
                    "There is no event with id `{}`. Check configuration for correct event id.".format(self.event_id))

            self.event.replace(event)

            profile_entity = Entity(id=self.event.profile.id)
            session_entity = Entity(id=self.event.session.id)
            profile_task = asyncio.create_task(profile_entity.storage('profile').load(Profile))
            session_task = asyncio.create_task(session_entity.storage('session').load(Session))

            profile = await profile_task
            session = await session_task

            if session is None:
                raise ValueError(
                    "Event id `{}` has reference to empty session id `{}`. Debug stopped. This event is corrupted.".format(
                        self.event_id, self.event.session.id))

            self.session.replace(session)

            if profile is None:
                raise ValueError(
                    "Event id `{}` has reference to empty profile id `{}`. Debug stopped. This event is corrupted.".format(
                        self.event_id, self.event.profile.id))

            self.profile.replace(profile)

            return Result(port="event", value=self.event.dict())

        return Result(port="event", value=self.event.dict())


def register() -> Plugin:
    return Plugin(
        start=True,
        debug=True,
        spec=Spec(
            module='tracardi.process_engine.action.v1.debug_payload_action',
            className='DebugPayloadAction',
            inputs=[],
            outputs=["event"],
            init={
                "event": {
                    "id": "undefined",
                }
            },
            version='0.1',
            license="MIT",
            author="Risto Kowaczewski"
        ),
        metadata=MetaData(
            name='Debug',
            desc='Loads debug payload into flow. This action is executed only in debug mode. ' +
                 'Use it to inject payload defined it config to analyse you workflow.',
            keywords=['start node'],
            type='flowNode',
            width=100,
            height=100,
            icon='debug',
            group=["Input/Output"]
        )
    )



def validate_list_values(values):
    for value in values:
        if type(value) not in [str, int, float, bool]:
            raise ValueError("Invalid value in list `{}`".format(values))


def append(base, key, value):
    if not isinstance(base, dict):
        raise ValueError("Could not merge value `{}` and values `{}`".format(
            base,
            {key: value}))

    # Treat tuple and set as list
    if key in base:
        if type(base[key]) in [tuple, set]:
            base[key] = list(base[key])
        if isinstance(base[key], list):
            validate_list_values(base[key])

    if type(value) in [set, tuple]:
        value = list(value)
    if isinstance(value, list):
        validate_list_values(value)

    # Merge

    if key not in base:
        base[key] = value
    else:
        """
        Existing value can be only a list or a simple type
        """

        # Existing value is a list
        if isinstance(base[key], list):
            if isinstance(value, list):
                base[key] += value
            else:
                base[key].append(value)
            # Existing value is not a dict and value is a simple value
        elif not isinstance(base[key], dict) and type(value) in [str, int, float, bool, list]:
            base[key] = [base[key]]
            if isinstance(value, list):
                validate_list_values(value)
                base[key] += value
            else:
                base[key].append(value)
        else:
            raise ValueError("Could not merge `{}` with value `{}`".format(base[key], value))

    # make uniq
    if isinstance(base[key], list):
        base[key] = list(set(base[key]))
        if len(base[key]) == 1:
            base[key] = base[key][0]

    return base


def merge(base, dict_list):
    for key in set().union(*dict_list):
        for data in dict_list:
            if key in data:
                # Null
                if data[key] is None:
                    continue
                # Simple types
                elif type(data[key]) in [str, int, float, bool, tuple, list, set]:
                    append(base, key, data[key])
                # Dicts
                elif type(data[key]) in [dict]:
                    if key not in base:
                        base[key] = {}
                    base[key] = merge(base[key], [data[key]])
                # Objects
                elif isinstance(data[key], object):
                    raise ValueError("Object of type `{}: {}` can not be merged with value `{}: {}`".format(
                        key, type(data[key]),
                        key, base[key] if key in base else base))
                else:
                    raise ValueError("Unknown type `{}: {}`".format(key, type(data[key])))

    return base



from enum import Enum


class IndexesSearch(str, Enum):
    source = "source"
    session = "session"
    profile = "profile"
    event = "event"
    rule = "rule"
    segment = "segment"
    flow = "flow"



import dateparser

from .transformer_namespace import TransformerNamespace


class CommonTransformer(TransformerNamespace):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def ESCAPED_STRING(self, args):
        return 'ESCAPED_STRING', args.value.replace("\"", "")

    def INTEGER(self, args):
        return 'INTEGER', args.value

    def BOOL(self, args):
        return 'BOOL', True if args.value.lower() == 'true' else False

    def FLOAT(self, args):
        return 'FLOAT', args.value

    def value(self, args):
        type = args[0].type

        # remove namespace
        if '__' in type:
            type = type.split('__')[-1]

        if type == "ESCAPED_STRING":
            value = args[0].value.replace("\"", "")
        else:
            value = args[0].value
        return type, value

    def array(self, args):
        return 'ARRAY', args

    def compound_parser(self, args):
        return args[0].value

    def compound_value(self, args):
        value = args[1][1]
        value_type = args[0].lower()
        if value_type == 'date':
            date = dateparser.parse(value)
            if not date:
                raise ValueError("Could not parse date `{}`".format(value))
            value = date.strftime("%Y-%m-%dT%H:%M:%SZ")
        else:
            raise ValueError("Unknown parse function `{}`".format(value_type))

        return value_type,value

    def DOTTED_FIELD(self, args):
        return 'DOTTED_FIELD', args.value


import elasticsearch
import pytest

import tracardi
from .mocks.mock_storage import MockStorageCrud
from tracardi.domain.entity import Entity
from tracardi.domain.profile import Profile
from tracardi.domain.session import Session
from tracardi.domain.payload.tracker_payload import TrackerPayload


@pytest.mark.asyncio
async def test_get_context_session_exists_profile_not_exists(mocker):

    mocker.patch('tracardi.domain.session.StorageCrud', MockStorageCrud)
    mocker.patch('tracardi.domain.session.StorageCrud', MockStorageCrud)
    mocker.patch('tracardi.domain.entity.crud', tracardi.tests.unit.mocks.mock_storage)
    mocker.patch('tracardi.domain.profile.StorageCrud', MockStorageCrud)

    # Session exists and has profile equal to 1
    tracker_payload = TrackerPayload(
        session=Entity(id="1"),
        profile=None,
        source=Entity(id="scope")
    )

    profile, session = await tracker_payload._get_profile_and_session()

    assert isinstance(profile, Profile)
    assert isinstance(session, Session)

    assert profile.id == "1"
    assert session.id == "1"


@pytest.mark.asyncio
async def test_get_context_session_exists_profile_exists_conflict(mocker):
    mocker.patch('tracardi.domain.session.StorageCrud', MockStorageCrud)
    mocker.patch('tracardi.domain.session.StorageCrud', MockStorageCrud)
    mocker.patch('tracardi.domain.entity.crud', tracardi.tests.unit.mocks.mock_storage)
    mocker.patch('tracardi.domain.profile.StorageCrud', MockStorageCrud)

    # Session exists and has profile equal to 1
    tracker_payload = TrackerPayload(
        session=Entity(id="1"),
        profile=Entity(id="2"),
        source=Entity(id="scope")
    )

    profile, session = await tracker_payload._get_profile_and_session()

    assert isinstance(profile, Profile)
    assert isinstance(session, Session)

    # profile from session is most important
    assert profile.id == "1"
    assert session.id == "1"


@pytest.mark.asyncio
async def test_get_context_session_not_exists_profile_exists(mocker):
    mocker.patch('tracardi.domain.session.StorageCrud', MockStorageCrud)
    mocker.patch('tracardi.domain.session.StorageCrud', MockStorageCrud)
    mocker.patch('tracardi.domain.entity.crud', tracardi.tests.unit.mocks.mock_storage)
    mocker.patch('tracardi.domain.profile.StorageCrud', MockStorageCrud)

    # Session 2 does not exist
    # Profile 2 exists
    tracker_payload = TrackerPayload(
        session=Entity(id="generated_not_exists-01"),
        profile=Entity(id="2"),
        source=Entity(id="scope")
    )

    profile, session = await tracker_payload._get_profile_and_session()

    assert isinstance(profile, Profile)
    assert isinstance(session, Session)

    # Do not generate profile copy it from param
    assert session.id == "generated_not_exists-01"
    assert profile.id == "2"


@pytest.mark.asyncio
async def test_get_context_session_not_exists_profile_not_exists(mocker):
    mocker.patch('tracardi.domain.session.StorageCrud', MockStorageCrud)
    mocker.patch('tracardi.domain.session.StorageCrud', MockStorageCrud)
    mocker.patch('tracardi.domain.entity.crud', tracardi.tests.unit.mocks.mock_storage)
    mocker.patch('tracardi.domain.profile.StorageCrud', MockStorageCrud)

    tracker_payload = TrackerPayload(
        session=Entity(id="generated_not_exists-02"),
        profile=None,
        source=Entity(id="scope")
    )

    profile, session = await tracker_payload._get_profile_and_session()

    assert isinstance(profile, Profile)
    assert isinstance(session, Session)

    # Profile generated
    assert profile.id != "2"
    # Session generated
    assert session.id == "generated_not_exists-02"


@pytest.mark.asyncio
async def test_load_session_with_error(mocker):
    def exc(*args):
        raise elasticsearch.exceptions.ElasticsearchException("test")

    mocker.patch('tracardi.domain.session.StorageCrud.load', exc)

    try:
        tracker_payload = TrackerPayload(
            session=Entity(id="non_existent"),
            profile=None,
            source=Entity(id="scope")
        )

        profile, session = await tracker_payload._get_profile_and_session()

    except elasticsearch.exceptions.ElasticsearchException as e:
        assert str(e) == "test"



from tracardi.event_server.service.persistence_service import PersistenceService
from tracardi.service.storage.collection_crud import CollectionCrud
from tracardi.service.storage.elastic_storage import ElasticStorage


class Segments(list):

    # Persistence
    def bulk(self) -> CollectionCrud:
        return CollectionCrud("segment", self)

    @staticmethod
    def storage() -> PersistenceService:
        return PersistenceService(ElasticStorage(index_key="segment"))



import asyncio

from tracardi.service.storage.elastic import Elastic
from tracardi.service.storage.sql import to_sql_query


async def main():
    es = Elastic.instance()
    q = to_sql_query("test1-tracardi-event", query="type='view' and event_server.browser.browser.engine='xxx'")
    print(q)
    q = await es.translate(q)
    print(q)
    await es.close()

loop = asyncio.get_event_loop()
loop.run_until_complete(main())
loop.close()


from tracardi.domain.session import Session
from tracardi.domain.value_object.operation import Operation


def test_session():
    session = Session(id='1')
    assert isinstance(session.operation, Operation)



import logging
from datetime import datetime

from tracardi.domain.query_result import QueryResult
from tracardi.domain.time_range_query import DatetimeRangePayload
from tracardi.event_server.service.persistence_service import PersistenceService
from tracardi.exceptions.exception import StorageException
from tracardi.service.storage import index as index_2_elastic
from tracardi.service.storage.elastic_storage import ElasticStorage
from tracardi.service.storage.sql import to_sql_query, to_time_range_sql_query

_logger = logging.getLogger("Index")


class Index:

    def __init__(self, index: str):
        self.index = index
        self.read_index = index_2_elastic.resources[index].get_read_index()
        self.storage_service = PersistenceService(ElasticStorage(index_key=self.index))
        self.time_fields_map = {
            'event': 'metadata.time.insert',
            'session': 'metadata.time.insert',
            'profile': 'metadata.time.insert',
        }

    async def search(self, query: str = None, limit: int = 20):
        query = to_sql_query(self.read_index, query=query, limit=limit)
        return (await self.storage_service.sql(query)).dict()

    @staticmethod
    def _convert_time_zone(query, min_date_time, max_date_time):
        time_zone = "UTC" if query.timeZone is None or query.timeZone == "" else query.timeZone

        if time_zone != "UTC":
            min_date_time, time_zone = query.convert_to_local_datetime(min_date_time, time_zone)
            max_date_time, time_zone = query.convert_to_local_datetime(max_date_time, time_zone)

        return min_date_time, max_date_time, time_zone

    async def time_range(self, query: DatetimeRangePayload) -> QueryResult:

        if self.index not in self.time_fields_map:
            raise ValueError("No time_field available on `{}`".format(self.index))

        min_date_time, max_date_time = query.get_dates()  # type: datetime, datetime
        min_date_time, max_date_time, time_zone = self._convert_time_zone(query, min_date_time, max_date_time)

        limit = query.limit
        offset = query.offset
        sql = query.where
        time_field = self.time_fields_map[self.index]

        sql = to_time_range_sql_query(self.read_index, time_field, min_date_time, max_date_time, sql)
        try:
            translated_query = await self.storage_service.translate(sql)
        except StorageException as e:
            _logger.error("Could not translate SQL {}. Reason: {}".format(sql, str(e)))
            return QueryResult(total=0, result=[])

        # Add time zone

        if time_zone:
            if 'range' in translated_query['query']:
                translated_query['query']['range'][time_field]['time_zone'] = time_zone
            elif 'bool' in translated_query['query'] and 'must' in translated_query['query']['bool']:
                translated_query['query']['bool']['must'][0]['range'][time_field]['time_zone'] = time_zone

        es_query = {
            "from": offset,
            "size": limit,
            "query": translated_query['query'],
            "sort": [
                {
                    time_field: "desc"
                }
            ]
        }

        try:
            result = await self.storage_service.filter(es_query)
        except StorageException as e:
            _logger.error("Could not filter {}. Reason: {}".format(es_query, str(e)))
            return QueryResult(total=0, result=[])

        return QueryResult(**result.dict())

    async def histogram(self, query: DatetimeRangePayload):

        def __interval(min: datetime, max: datetime):

            max_interval = 50
            min_interval = 20

            span = max - min

            if span.days > max_interval:
                # up
                return int(span.days / max_interval), 'd', "%y-%m-%d"
            elif min_interval > span.days:
                # down
                interval = int((span.days * 24) / max_interval)
                if interval > 0:
                    return interval, 'h', "%d/%m %H:%M"

                # minutes
                interval = int((span.days * 24 * 60) / max_interval)
                if interval > 0:
                    return interval, 'm', "%H:%M"

                return 1, 'm', "%H:%M"

            return 1, 'd', "%y-%m-%d"

        def __format(data, unit, interval, format):
            for row in data:
                # todo timestamp no timezone
                timestamp = datetime.fromisoformat(row["key_as_string"].replace('Z', '+00:00'))
                yield {
                    "time": "{}".format(timestamp.strftime(format)),
                    'interval': "+{}{}".format(interval, unit),
                    "events": row["doc_count"]
                }

        min_date_time, max_date_time = query.get_dates()  # type: datetime, datetime
        min_date_time, max_date_time, time_zone = self._convert_time_zone(query, min_date_time, max_date_time)

        # limit = query.limit
        # offset = query.offset
        sql = query.where
        time_field = self.time_fields_map[self.index]

        interval, unit, format = __interval(min_date_time, max_date_time)

        sql = to_time_range_sql_query(self.read_index, time_field, min_date_time, max_date_time, sql)

        try:
            translated_query = await self.storage_service.translate(sql)
        except StorageException as e:
            _logger.error("Could not translate SQL {}. Reason: {}".format(sql, str(e)))
            return QueryResult(total=0, result=[])

        if time_zone:
            if 'range' in translated_query['query']:
                translated_query['query']['range'][time_field]['time_zone'] = time_zone
            elif 'bool' in translated_query['query'] and 'must' in translated_query['query']['bool']:
                translated_query['query']['bool']['must'][0]['range'][time_field]['time_zone'] = time_zone

        es_query = {
            "size": 0,
            "query": translated_query['query'],
            "aggs": {
                "items_over_time": {
                    "date_histogram": {
                        "min_doc_count": 0,
                        "field": time_field,
                        "fixed_interval": f"{interval}{unit}",
                        "extended_bounds": {
                            "min": min_date_time,
                            "max": max_date_time
                        }
                    }
                }
            }
        }

        if time_zone:
            es_query['aggs']['items_over_time']['date_histogram']['time_zone'] = time_zone

        try:
            result = await self.storage_service.query(es_query)
        except StorageException as e:
            _logger.error("Could not query {}. Reason: {}".format(es_query, str(e)))
            return QueryResult(total=0, result=[])

        qs = {
            'total': result['hits']['total']['value'],
            'result': list(__format(result['aggregations']['items_over_time']['buckets'], unit, interval, format))
        }

        return QueryResult(**qs)



from tracardi_plugin_sdk.domain.register import Plugin, Spec, MetaData
from tracardi_plugin_sdk.action_runner import ActionRunner
from tracardi_plugin_sdk.domain.result import Result
from tracardi.domain.profile import Profile
from tracardi_dot_notation.dot_accessor import DotAccessor


class DecrementAction(ActionRunner):

    def __init__(self, **kwargs):

        if 'field' not in kwargs or kwargs['field'] is None:
            raise ValueError("Field is not set. Define it in config section.")

        if 'decrement' not in kwargs or kwargs['decrement'] is None:
            raise ValueError("Decrement is not set. Define it in config section.")

        self.field = kwargs['field']
        self.decrement = kwargs['decrement']

        if type(self.decrement) != int and type(self.decrement) != float:
            raise ValueError("Decrement must be a number. {} given.".format(type(self.decrement)))

        if not self.field.startswith('profile@stats.counters'):
            raise ValueError("Only fields inside `profile@stats.counters` can be decremented. Field `{}` given.".format(self.field))

    async def run(self, payload):

        dot = DotAccessor(
            self.profile,
            self.session,
            payload if isinstance(payload, dict) else None,
            self.event,
            self.flow)

        try:

            value = dot[self.field]

            if value is None:
                value = 0

        except KeyError:
            value = 0

        if type(value) != int:
            raise ValueError("Filed `{}` value is not numeric.".format(self.field))

        value -= self.decrement

        dot[self.field] = value

        self.profile.replace(Profile(**dot.profile))

        return Result(port="payload", value=payload)


def register() -> Plugin:
    return Plugin(
        start=False,
        spec=Spec(
            module='tracardi.process_engine.action.v1.decrement_action',
            className='DecrementAction',
            inputs=["payload"],
            outputs=['payload'],
            init={"field": None, "decrement": 1},
            manual="decrement_action",
            version='0.1',
            license="MIT",
            author="Risto Kowaczewski"
        ),
        metadata=MetaData(
            name='Decrement counter',
            desc='Decrement profile stats.counters value. Returns payload',
            type='flowNode',
            width=200,
            height=100,
            icon='minus',
            group=["Stats"]
        )
    )




import pytest

from tracardi.domain.entity import Entity
from tracardi.domain.payload.event_payload import EventPayload
from tracardi.domain.payload.tracker_payload import TrackerPayload


@pytest.mark.asyncio
async def test_tracker_collect():

    source = Entity(id="scope")
    await source.storage("source").save({})

    session = Entity.new()
    profile = Entity.new()

    tracker_payload = TrackerPayload(
        session=session,
        profile=profile,
        source=source,
        events=[
            EventPayload(
                type="click",
                properties={"btn": [1, 2, 3]},
                options={"save": False}
            ),
            EventPayload(
                type="click",
                properties={"btn": [3, 4, 5]},
                options={"save": True}
            )
        ]
    )

    profile, session, events, result = await tracker_payload.collect()

    # assert profile.id == result.profile.ids[0]
    assert session.id == result.session.ids[0]

    assert result.session.saved == 1
    # assert result.profile.saved == 1
    assert result.events.saved == 1  # Only one event is marked to saved

    for id in result.events.ids:
        entity = Entity(id=id)
        await entity.storage("event").delete()

    # for id in result.profile.ids:
    #     entity = Entity(id=id)
    #     await entity.storage("profile").delete()

    for id in result.session.ids:
        entity = Entity(id=id)
        await entity.storage("session").delete()






import copy
import uuid
from datetime import datetime
from typing import Optional, Any, List, Union

from .entity import Entity
from .metadata import Metadata
from .pii import PII
from .profile_traits import ProfileTraits
from .time import Time
from .value_object.operation import Operation
from ..service.dot_notation_converter import DotNotationConverter
from ..service.storage.collection_crud import CollectionCrud
from ..service.storage.crud import StorageCrud

from .profile_stats import ProfileStats
from ..service.merger import merge
from .segment import Segment
from .segments import Segments
from ..process_engine.tql.condition import Condition
from ..process_engine.tql.utils.dictonary import flatten


class Profile(Entity):
    mergedWith: Optional[str] = None
    metadata: Optional[Metadata]
    operation: Optional[Operation] = Operation()
    stats: ProfileStats = ProfileStats()
    traits: Optional[ProfileTraits] = ProfileTraits()
    pii: PII = PII()
    segments: Optional[list] = []
    consents: Optional[dict] = {}
    active: bool = True

    def __init__(self, **data: Any):
        data['metadata'] = Metadata(
            time=Time(
                insert=datetime.utcnow()
            ))
        super().__init__(**data)

    def replace(self, profile):
        self.metadata = profile.metadata
        self.stats = profile.stats
        self.traits = profile.traits
        self.pii = profile.pii
        self.id = profile.id
        self.segments = profile.segments
        self.consents = profile.consents
        self.active = profile.active
        self.mergedWith = profile.mergedWith
        self.operation = profile.operation

    def get_merge_key_values(self) -> List[tuple]:
        converter = DotNotationConverter(self)
        return [converter.get_profile_fiel_value_pair(key) for key in self.operation.merge]

    def _get_merging_keys_and_values(self):
        merge_key_values = self.get_merge_key_values()

        # Add keyword
        merge_key_values = [(f"{field}.keyword", value) for field, value in merge_key_values if value is not None]

        return merge_key_values

    @staticmethod
    def _mark_profiles_as_merged(profiles, merge_with) -> List['Profile']:
        disabled_profiles = []

        for profile in profiles:
            profile.active = False
            profile.mergedWith = merge_with
            disabled_profiles.append(profile)

        return disabled_profiles

    async def segment(self, event_types):

        """
        This method mutates current profile. Loads segments and adds segments to current profile.
        """

        # todo cache segments for 30 sec
        flat_payload = flatten(copy.deepcopy(self.dict()))

        for event_type in event_types:  # type: str

            # Segmentation is run for every event

            # todo segments are loaded one by one - maybe it is possible to load it at once
            segments = await Segments.storage().load_by('eventType.keyword', event_type)
            for segment in segments:

                segment = Segment(**segment)
                segment_id = segment.get_id()

                try:

                    if Condition.evaluate(segment.condition, flat_payload):
                        segments = set(self.segments)
                        segments.add(segment_id)
                        self.segments = list(segments)

                        # Yield only if segmentation triggered
                        yield event_type, segment_id, None

                except Exception as e:
                    msg = 'Condition id `{}` could not evaluate `{}`. The following error was raised: `{}`'.format(
                        segment_id, segment.condition, str(e).replace("\n", " "))

                    yield event_type, segment_id, msg

    async def merge(self, limit=2000) -> Union['Profiles', None]:

        """
        This method mutates current profile.
        Merges profiles on keys set in profile.operation.merge. Loads profiles from database and
        combines its data into current profile. Returns Profiles object with profiles to be disables.
        It does not disable profiles or saves merged profile.
        """

        merge_key_values = self._get_merging_keys_and_values()

        # Are there any non-empty values in current profile

        if len(merge_key_values) > 0:

            # Load all profiles that match merging criteria
            existing_profiles = await tracardi.domain.value_object.operation.Operation.load_profiles_to_merge(
                merge_key_values,
                limit=limit)

            # Filter only profiles that are not current profile and where not merged
            profiles_to_merge = [p for p in existing_profiles if p.id != self.id and p.active is True]

            print('profiles_to_merge',profiles_to_merge)

            # Are there any profiles to merge?
            if len(profiles_to_merge) > 0:
                # Add current profile to existing ones and get merged profile
                merged_profile = Profiles.merge(profiles_to_merge, self)

                # Replace current profile with merged profile
                self.replace(merged_profile)

                # Deactivate all other profiles except merged one

                profiles_to_disable = [p for p in existing_profiles if p.id != self.id]
                disabled_profiles = self._mark_profiles_as_merged(profiles_to_disable, merge_with=self.id)

                return Profiles(disabled_profiles)

        return None

    def increase_visits(self, value=1):
        self.stats.visits += value

    def increase_views(self, value=1):
        self.stats.views += value

    def storage(self) -> StorageCrud:
        return StorageCrud("profile", Profile, entity=self, exclude={"operation": ...})

    @staticmethod
    async def load_current(id) -> 'Profile':

        """
        Loads current profile. If profile was merged then it loads merged profile.
        """

        entity = Entity(id=id)
        profile = await entity.storage('profile').load(Profile)  # type: Profile
        if profile is not None and profile.mergedWith is not None:
            profile = await Profile.load_current(profile.mergedWith)
        return profile

    @staticmethod
    def new() -> 'Profile':
        """
        @return Profile
        """
        return Profile(id=str(uuid.uuid4()))


class Profiles(list):

    @staticmethod
    def merge(existing_profiles: List[Profile], current_profile: Profile) -> Profile:

        profiles = existing_profiles + [current_profile]

        traits = [profile.traits.dict() for profile in profiles]
        traits = merge({}, traits)

        piis = [profile.pii.dict() for profile in profiles]
        piis = merge({}, piis)

        consents = {}
        segments = []
        stats = ProfileStats()
        for profile in profiles:

            stats.visits += profile.stats.visits
            stats.views += profile.stats.views

            if isinstance(profile.segments, list):
                segments += profile.segments

            consents.update(profile.consents)

            # make uniq
            segments = list(set(segments))

        # Set id to merged id or current profile id.
        id = current_profile.mergedWith if current_profile.mergedWith is not None else current_profile.id

        return Profile(
            id=id,
            mergedWith=None,
            stats=stats,
            traits=traits,
            pii=piis,
            segments=segments,
            consents=consents,
            active=True
        )

    def bulk(self) -> CollectionCrud:
        return CollectionCrud("profile", self)



from tracardi.domain.entity import Entity
from tracardi.service.storage.collection_crud import CollectionCrud
from tracardi.service.storage.crud import StorageCrud


class Project(Entity):
    name: str

    def storage(self) -> StorageCrud:
        return StorageCrud("project", Project, entity=self)


class Projects(list):

    def bulk(self) -> CollectionCrud:
        return CollectionCrud("project", self)



from asyncio import sleep

from tracardi_plugin_sdk.domain.register import Plugin, Spec, MetaData
from tracardi_plugin_sdk.action_runner import ActionRunner
from tracardi_plugin_sdk.domain.result import Result


class SleepAction(ActionRunner):

    def __init__(self, **kwargs):
        self.sleep = kwargs['sleep'] if 'sleep' in kwargs else 0

    async def run(self, payload):
        await sleep(self.sleep)
        return Result(port="payload", value=payload)


def register() -> Plugin:
    return Plugin(
        start=False,
        spec=Spec(
            module='tracardi.process_engine.action.v1.sleep_action',
            className='SleepAction',
            inputs=["payload"],
            outputs=["payload"],
            version='0.1.1',
            license="MIT",
            author="Risto Kowaczewski",
            init={"sleep": 1}

        ),
        metadata=MetaData(
            name='Sleep',
            desc='Sleeps workflow.',
            type='flowNode',
            width=100,
            height=100,
            icon='sleep',
            group=["Time"]
        )
    )



from datetime import datetime
from typing import Optional, List, Tuple, Any
from pydantic import BaseModel

from ..value_object.collect_result import CollectResult
from ..entity import Entity
from ..events import Events
from ..payload.event_payload import EventPayload
from ..profile import Profile
from ..session import Session
from ..metadata import Metadata
from ..time import Time
from ..value_object.bulk_insert_result import BulkInsertResult
from ..value_object.save_result import SaveResult
from ..value_object.tracker_payload_result import TrackerPayloadResult
from ...config import tracardi
from ...process_engine.rules_engine import RulesEngine


class TrackerPayload(BaseModel):
    source: Entity
    session: Entity

    metadata: Optional[Metadata]
    profile: Optional[Entity] = None
    context: Optional[dict] = {}
    properties: Optional[dict] = {}
    events: List[EventPayload] = []
    options: Optional[dict] = {}

    def __init__(self, **data: Any):
        data['metadata'] = Metadata(
            time=Time(
                insert=datetime.utcnow()
            ))
        super().__init__(**data)

    def get_events(self, session: Session, profile: Profile) -> Events:
        _events = Events()
        if self.events:
            for event in self.events:  # type: EventPayload
                if event.is_persistent():
                    _event = event.to_event(self.metadata, self.source, session, profile, self.options)
                    _events.append(_event)
        return _events

    def return_profile(self):
        return self.options and "profile" in self.options and self.options['profile'] is True

    def is_disabled(self, key):
        return key in self.options and self.options[key] is False

    async def collect(self) -> Tuple[Profile, Session, Events, CollectResult]:

        save_session = False if self.is_disabled('saveSession') else True
        save_events = False if self.is_disabled('saveEvents') else True
        save_profile = True

        # Get profile, session objects
        profile, session = await self._get_profile_and_session()

        # Presistence

        # Save profile
        if save_profile and profile.operation.new:
            profile_result = await profile.storage().save()
        else:
            profile_result = BulkInsertResult()

        # Save session
        if save_session and (session.operation.new or session.profile is None or session.profile.id != profile.id):
            # save only profile Entity
            session.profile = Entity(id=profile.id)
            session_result = await session.storage().save()
        else:
            session_result = BulkInsertResult()

        # Save events
        if save_events:
            events = self.get_events(session, profile)
            event_result = await events.bulk().save()
            event_result = SaveResult(**event_result.dict())

            # Add event types
            for event in events:
                event_result.types.append(event.type)
        else:
            events = Events()
            event_result = BulkInsertResult()

        result = {
            "session": session_result,
            "events": event_result,
            "profile": profile_result

        }
        return profile, session, events, CollectResult(**result)

    async def _get_profile_and_session(self) -> Tuple[Profile, Session]:

        """
        Returns session. Creates profile if it does not exist.If it exists connects session with profile.
        """

        if self.session.id is None:
            raise ValueError("Session must be set.")

        # Load session from storage
        if isinstance(self.session, Entity):
            session = await self.session.storage("session").load(Session) # type: Session
        else:
            raise ValueError(
                "Session has unknown type. Available types Entity, Session. `{}` given".format(type(self.session)))

        is_new_profile = False
        is_new_session = False

        if session is None:

            session = Session(id=self.session.id)
            is_new_session = True

            # Bind profile

            if self.profile is None:

                # Create empty default profile generate profile_id
                profile = Profile.new()

                # Create new profile
                is_new_profile = True

            else:

                # Id exists load profile from storage
                profile = await Profile.load_current(id=self.profile.id)  # type: Profile

                if profile is None:
                    # Profile id delivered but profile does not exist in storage.
                    # Id was forged

                    profile = Profile.new()

                    # Create new profile
                    is_new_profile = True

        else:

            # There is session. Copy profile id form session to profile

            if session.profile is not None:
                # Loaded session has profile

                # Load profile based on profile id saved in session
                profile = await Profile.load_current(id=session.profile.id)  # type: Profile

                if session.profile.id != profile.id:
                    # Profile in session id has been merged. Change profile in session.

                    session.profile.id = profile.id
                    session.metadata.time = Time(insert=datetime.utcnow())

                    is_new_session = True

                # profile = await profile.storage().load()  # type: Profile

            else:
                # Corrupted session

                profile = None

            if profile is None:
                # Id exists but profile not exist in storage.

                profile = Profile.new()

                # Create new profile
                is_new_profile = True

        session.context = self.context
        session.properties = self.properties
        session.operation.new = is_new_session

        profile.operation.new = is_new_profile

        return profile, session

    async def process(self) -> Tuple[dict, Profile]:

        profile, session, events, collect_result = await self.collect()

        rules_engine = RulesEngine(
            session,
            profile,
            events)

        debug_info_by_event_type_and_rule_name, segmentation_result, console_log = await rules_engine.execute(
            self.source.id
        )

        # Prepare response
        result = {}

        # Debugging
        # todo save result to different index
        if not tracardi.track_debug:
            debug_result = TrackerPayloadResult(**collect_result.dict())
            debug_result = debug_result.dict()
            debug_result['execution'] = debug_info_by_event_type_and_rule_name
            debug_result['segmentation'] = segmentation_result
            debug_result['logs'] = console_log
            result['debugging'] = debug_result

        # Profile
        if self.return_profile():
            result["profile"] = profile.dict(
                exclude={
                    "traits": {"private": ...},
                    "pii": ...,
                    "operation": ...
                })
        else:
            result["profile"] = profile.dict(include={"id": ...})

        return result, profile



from datetime import datetime


class Index:
    def __init__(self, multi_index, index, mapping, rel):
        self.multi_index = multi_index
        self.rel = rel
        self.index = index
        self.mapping = mapping

    def get_read_index(self):
        if self.multi_index is False:
            return self.index

        return self.index + "*"

    def get_write_index(self):
        if self.multi_index is False:
            return self.index

        date = datetime.now()
        return self.index + f"-{date.year}-{date.month}"


class Resource:

    def __init__(self):
        self.resources = {
            # "credential":   Index(multi_index=False, index="tracardi-credential",          mapping=None, rel=None),
            "project":      Index(multi_index=False, index="tracardi-flow-project",        mapping=None, rel=None),
            "action":       Index(multi_index=False, index="tracardi-flow-action-plugins", mapping=None, rel=None),
            "token":        Index(multi_index=False, index="tracardi-token",               mapping="mappings/token-index.json", rel=None),
            "source":       Index(multi_index=False, index="tracardi-source",              mapping=None, rel=None),
            "session":      Index(multi_index=True,  index="tracardi-session",             mapping=None, rel='profile.id'),
            "profile":      Index(multi_index=False, index="tracardi-profile",             mapping="mappings/profile-index.json", rel='_id'),
            "event":        Index(multi_index=True,  index="tracardi-event",               mapping="mappings/event-index.json", rel=None),
            "flow":         Index(multi_index=False, index="tracardi-flow",                mapping=None, rel=None),
            "rule":         Index(multi_index=False, index="tracardi-rule",                mapping=None, rel=None),
            "segment":      Index(multi_index=False, index="tracardi-segment",             mapping=None, rel=None),
            # "consent":      Index(multi_index=False, index="tracardi-consent",             mapping=None, rel=None),
            "console-log":  Index(multi_index=False, index="tracardi-console-log",         mapping="mappings/console-log-index.json", rel=None),
            "stat-log":     Index(multi_index=False, index="tracardi-stat-log",            mapping=None, rel=None),
            "debug-info":   Index(multi_index=False, index="tracardi-debug-info",          mapping="mappings/debug-info-index.json", rel=None),
            # "response":     Index(multi_index=False, index="tracardi-response",            mapping=None, rel=None),
        }

    def __getitem__(self, item):
        return self.resources[item]

    def __contains__(self, item):
        return item in self.resources


resources = Resource()



import dateparser
from tracardi_dot_notation.dot_accessor import DotAccessor

from ..domain.field import Field
from .function_transformer import FunctionTransformer
from .transformer_namespace import TransformerNamespace

operation_mapper = {
    "between": "between",
    "=": "equals",
    "!=": "notEquals",
    "<>": "notEquals",  # todo implement in parser
    '>=': 'greaterThanOrEqualTo',
    '<=': 'lessThanOrEqualTo',
    '>': 'greaterThan',
    '<': 'lessThan',
    'is null': 'isNull',
    'startsWith': 'starts with',  # todo: implement,
    'endsWith': 'ends with',  # todo: implement,
    'matchesRegex': 'regex',  # todo: implement,
    'in': 'in',  # todo: implement,
    'not in': 'not in',  # todo: implement,
    'isDay': 'is day',  # todo: implement,
    'isNotDay': 'is not day',  # todo: implement,
}


class ExprTransformer(TransformerNamespace):

    def __init__(self, dot, *args, **kwargs):
        if not isinstance(dot, DotAccessor):
            raise ValueError("Data passed to ExprTransformer must be type of DotAccessor.")
        super().__init__(*args, **kwargs)
        self.namespace('uql_function__', FunctionTransformer())
        self._dot = dot

    def expr(self, args):
        return args[0]

    def and_expr(self, args):
        # return args
        value1, _, value2 = args
        return value1 and value2

    def or_expr(self, args):
        # return args
        value1, _, value2 = args
        return value1 or value2

    def OP_FIELD(self, args):
        return Field(args.value, self._dot)

    def OP(self, args):
        return args.value

    def OP_INTEGER(self, args):
        return int(args.value)

    def OP_STRING(self, args):
        return args.value.strip('"')

    @staticmethod
    def _compare(operation, value1, value2):
        if operation == '==':
            if isinstance(value1, list) and not isinstance(value2, list):
                return value2 in value1
            return value1 == value2
        elif operation == '!=':
            return value1 != value2
        elif operation == '>':
            return value1 > value2
        elif operation == '>=':
            return value1 >= value2
        elif operation == '<':
            return value1 < value2
        elif operation == '<=':
            return value1 <= value2

    def op_condition(self, args):
        value1, operation, value2 = args
        return self._compare(operation, value1, value2)

    def op_array(self, args):
        return args

    def OP_BOOL(self, args):
        return args.lower() == 'true'

    def OP_NULL(self, args):
        return None

    def OP_FLOAT(self, args):
        return args

    def op_range(self, args):
        return args

    def op_between(self, args):
        field, _, values = args
        value1, value2 = values
        return value1 <= field <= value2

    def op_field_eq_field(self, args):
        value1, operation, value2 = args
        return self._compare(operation, value1, value2)

    def OP_VALUE_TYPE(self, args):
        return args.value

    def op_compound_value(self, args):
        value_type, value = args
        if value_type == 'datetime':

            if not isinstance(value, str):
                raise ValueError(
                    "Value of `{}` must be string to compare it with datetime. Type of {} given".format(value,
                                                                                                        type(value)))

            date = dateparser.parse(value)

            if not date:
                raise ValueError("Could not parse date `{}`".format(value))
            return date
        raise ValueError("Unknown type `{}`".format(value_type))

    def op_compound_field(self, args):
        value_type, field = args
        value = field.value
        if value_type == 'datetime':

            if not isinstance(value, str):
                raise ValueError(
                    "Value of `{}` must be string to compare it with datetime. Type of {} given".format(field.label,
                                                                                                        type(value)))

            date = dateparser.parse(value)
            if not date:
                raise ValueError("Could not parse date `{}`".format(value))
            return date

        raise ValueError("Unknown type `{}`".format(value_type))

    def op_field_sig(self, args):
        return args[0]

    def op_value_sig(self, args):
        return args[0]

    def op_is_null(self, args):
        return args[0].value is None

    def op_exists(self, args):
        return args[0].label in self._dot

    def op_not_exists(self, args):
        return args[0].label not in self._dot



from datetime import datetime
from typing import Optional, Any
from tracardi.service.storage.crud import StorageCrud
from ..entity import Entity
from ..flow_action_plugin import FlowActionPlugin
from tracardi.domain.flow import PluginRecord
from ..metadata import Metadata
from ..settings import Settings
from ..time import Time


class FlowActionPluginRecord(Entity):
    metadata: Optional[Metadata]
    plugin: PluginRecord
    settings: Optional[Settings] = Settings()

    def __init__(self, **data: Any):
        data['metadata'] = Metadata(
            time=Time(
                insert=datetime.utcnow()
            ))
        super().__init__(**data)

    # Persistence

    def storage(self) -> StorageCrud:
        return StorageCrud("action", FlowActionPluginRecord, entity=self)

    @staticmethod
    def encode(action: FlowActionPlugin) -> 'FlowActionPluginRecord':
        return FlowActionPluginRecord(
            id=action.id,
            plugin=PluginRecord.encode(action.plugin),
            metadata=action.metadata,
            settings=action.settings
        )

    def decode(self) -> FlowActionPlugin:
        data = {
            "id": self.id,
            "plugin": self.plugin.decode(),
            "metadata": self.metadata,
            "settings": self.settings
        }
        return FlowActionPlugin.construct(_fields_set=self.__fields_set__, **data)



class AggResult:
    def __init__(self, agg_key, result=None, return_counts=True):
        self.return_counts = return_counts
        if result is None:
            self.total = 0
            self._hits = []
        else:
            self.total = result['hits']['total']['value']
            self._hits = result['aggregations'][agg_key]['buckets']

    def __repr__(self):
        return "hits {}, total: {}".format(self._hits, self.total)

    def __iter__(self):
        for hit in self._hits:
            row = hit['key']
            if self.return_counts:
                count = hit['doc_count']
                yield {row: count}
            else:
                yield row

    def dict(self):
        return {
            "total": self.total,
            "result": list(self)
        }



import uuid
from tracardi_graph_runner.domain.flow import Flow as GraphFlow
import tracardi.service.storage.crud as crud
from .entity import Entity
from .named_entity import NamedEntity
from ..exceptions.exception import TracardiException
from typing import Optional, List
from pydantic import BaseModel
from tracardi_graph_runner.domain.flow_graph_data import FlowGraphData, Edge, Position, Node
from tracardi_plugin_sdk.domain.register import MetaData, Plugin, Spec
from tracardi.service.storage.crud import StorageCrud
from ..service.secrets import decrypt, encrypt


class Flow(GraphFlow):
    projects: Optional[List[str]] = ["General"]
    draft: Optional[str] = ""
    lock: bool = False

    # Persistence

    def storage(self) -> crud.StorageCrud:
        return crud.StorageCrud("flow", Flow, entity=self)

    @staticmethod
    async def decode(flow_id) -> 'Flow':
        flow_record_entity = Entity(id=flow_id)
        flow_record = await flow_record_entity.storage("flow").load(FlowRecord)  # type: FlowRecord

        if not flow_record:
            raise TracardiException("Could not find flow `{}`".format(flow_id))

        return flow_record.decode()

    def encode_draft(self, draft: 'Flow'):
        self.draft = encrypt(draft.dict())

    def decode_draft(self) -> 'Flow':
        flow = decrypt(self.draft)
        return Flow.construct(_fields_set=self.__fields_set__, **flow)

    @staticmethod
    def new(id:str = None) -> 'Flow':
        return Flow(
            id=str(uuid.uuid4()) if id is None else id,
            name="Empty",
            enabled=False,
            flowGraph=FlowGraphData(nodes=[], edges=[])
        )


class SpecRecord(BaseModel):
    className: str
    module: str
    inputs: Optional[List[str]] = []
    outputs: Optional[List[str]] = []
    init: Optional[str] = ""
    manual: Optional[str] = None
    author: Optional[str] = None
    license: Optional[str] = "MIT"
    version: Optional[str] = '0.0.1'

    @staticmethod
    def encode(spec: Spec) -> 'SpecRecord':
        return SpecRecord(
            className=spec.className,
            module=spec.module,
            inputs=spec.inputs,
            outputs=spec.outputs,
            init=encrypt(spec.init),
            manual=spec.manual,
            author=spec.author,
            license=spec.license,
            version=spec.version
        )

    def decode(self) -> Spec:
        return Spec(
            className=self.className,
            module=self.module,
            inputs=self.inputs,
            outputs=self.outputs,
            init=decrypt(self.init),
            manual=self.manual,
            author=self.author,
            license=self.license,
            version=self.version
        )


class PluginRecord(BaseModel):
    start: bool = False
    debug: bool = False
    spec: SpecRecord
    metadata: MetaData

    @staticmethod
    def encode(plugin: Plugin) -> 'PluginRecord':
        return PluginRecord(
            start=plugin.start,
            debug=plugin.debug,
            spec=SpecRecord.encode(plugin.spec),
            metadata=plugin.metadata
        )

    def decode(self) -> Plugin:
        data = {
            "start": self.start,
            "debug": self.debug,
            "spec": self.spec.decode(),
            "metadata": self.metadata
        }
        return Plugin.construct(_fields_set=self.__fields_set__, **data)


class NodeRecord(BaseModel):
    id: str
    type: str
    position: Position
    data: PluginRecord

    @staticmethod
    def encode(node: Node):
        return NodeRecord(
            id=node.id,
            type=node.type,
            position=node.position,
            data=PluginRecord.encode(node.data)
        )

    def decode(self) -> Node:
        data = {
            "id": self.id,
            "type": self.type,
            "data": self.data.decode(),
            "position": self.position
        }
        return Node.construct(_fields_set=self.__fields_set__, **data)


class FlowGraphDataRecord(BaseModel):
    nodes: List[NodeRecord]
    edges: List[Edge]

    @staticmethod
    def encode(flowGraph: FlowGraphData) -> 'FlowGraphDataRecord':
        if flowGraph:
            return FlowGraphDataRecord(
                edges=flowGraph.edges,
                nodes=[NodeRecord.encode(node) for node in flowGraph.nodes]
            )

        return FlowGraphDataRecord(
            edges=[],
            nodes=[]
        )

    def decode(self) -> FlowGraphData:
        data = {
            "edges": self.edges,
            "nodes": [node.decode() for node in self.nodes],
        }
        return FlowGraphData.construct(_fields_set=self.__fields_set__, **data)


class FlowRecord(NamedEntity):
    description: Optional[str] = None
    flowGraph: Optional[FlowGraphDataRecord] = None
    enabled: Optional[bool] = True
    projects: Optional[List[str]] = ["General"]
    draft: Optional[str] = ''
    lock: bool = False

    # Persistence

    def storage(self) -> StorageCrud:
        return StorageCrud("flow", FlowRecord, entity=self)

    @staticmethod
    def encode(flow: Flow) -> 'FlowRecord':
        return FlowRecord(
            id=flow.id,
            description=flow.description,
            name=flow.name,
            enabled=flow.enabled,
            flowGraph=FlowGraphDataRecord.encode(flow.flowGraph),
            projects=flow.projects,
            draft=flow.draft,
            lock=flow.lock

        )

    def decode(self) -> Flow:
        data = {
            "id": self.id,
            "name": self.name,
            "description": self.description,
            "enabled": self.enabled,
            "projects": self.projects,
            "draft": self.draft,
            "lock": self.lock,
            "flowGraph": self.flowGraph.decode() if self.flowGraph else None,
        }
        return Flow.construct(_fields_set=self.__fields_set__, **data)

    def decode_draft(self) -> 'Flow':
        flow = decrypt(self.draft)
        return Flow(**flow)

    def encode_draft(self, draft: 'Flow'):
        self.draft = encrypt(draft.dict())



from tracardi_plugin_sdk.domain.register import Plugin, Spec, MetaData
from tracardi_plugin_sdk.action_runner import ActionRunner


class UpdateProfileAction(ActionRunner):

    def __init__(self, *args, **kwargs):
        pass

    async def run(self, void):
        self.profile.operation.update = True
        return None


def register() -> Plugin:
    return Plugin(
        start=False,
        spec=Spec(
            module='tracardi.process_engine.action.v1.operations.update_profile_action',
            className='UpdateProfileAction',
            inputs=["void"],
            outputs=[],
            init=None,
            manual="update_profile_action"
        ),
        metadata=MetaData(
            name='Update profile',
            desc='Updates profile in storage.',
            type='flowNode',
            width=200,
            height=100,
            icon='store',
            group=["Operations"]
        )
    )



from tracardi_plugin_sdk.domain.register import Plugin, Spec, MetaData
from tracardi_plugin_sdk.action_runner import ActionRunner
from tracardi_plugin_sdk.domain.result import Result


class NewProfileAction(ActionRunner):

    def __init__(self, *args, **kwargs):
        pass

    async def run(self, void):
        if self.profile.operation.new:
            return Result(port="TRUE", value=True)

        return Result(port="FALSE", value=False)


def register() -> Plugin:
    return Plugin(
        start=False,
        spec=Spec(
            module='tracardi.process_engine.action.v1.new_profile_action',
            className='NewProfileAction',
            inputs=["void"],
            outputs=['TRUE', 'FALSE'],
            init=None,
            manual="new_profile_action",
            version='0.1',
            license="MIT",
            author="Risto Kowaczewski"
        ),
        metadata=MetaData(
            name='Is it a new profile',
            desc='If new profile then it returns true on TRUE output, otherwise returns false on FALSE port.',
            keywords=['condition'],
            type='flowNode',
            width=200,
            height=100,
            icon='question',
            group=["Conditions"]
        )
    )



from tracardi.domain.context import Context
from tracardi.domain.event import Event
from tracardi.domain.flow import Flow
from tracardi.domain.profile import Profile
from tracardi.domain.session import Session
from tracardi.domain.source import Source
from tracardi_dot_notation.dot_accessor import DotAccessor


def test_value_read():
    profile = Profile(id="1")
    session = Session(id="2")
    payload = {"a": 3}
    source = Source(id="3", type="event")
    context = Context()
    event = Event(id="event-id", type="type", source=source, context=context, profile=profile, session=session)
    flow = Flow(id="flow-id", name="flow")
    dot = DotAccessor(profile, session, payload, event, flow)

    assert dot['profile@id'] == "1"
    assert dot['session@id'] == "2"
    assert dot['payload@a'] == 3
    assert dot['flow@id'] == "flow-id"
    assert dot['event@id'] == "event-id"

    try:
        _ = dot['profile@none']
        assert False
    except KeyError:
        assert True

    dot['payload@b'] = 2
    assert dot['payload@b'] == 2

    dot['profile@other.a'] = 2
    assert dot['profile@other.a'] == 2

    assert dot.profile['other']['a'] == 2


def test_value_exists():
    profile = Profile(id="1")
    session = Session(id="2")
    payload = {"a": 3}
    source = Source(id="3", type="event")
    context = Context()
    event = Event(id="event-id", type="type", source=source, context=context, profile=profile, session=session)
    flow = Flow(id="flow-id", name="flow")
    dot = DotAccessor(profile, session, payload, event, flow)

    assert 'profile@id' in dot
    assert 'profile@missing' not in dot


def test_no_source():
    profile = Profile(id="1")
    session = Session(id="2")
    payload = {"a": 3}
    source = Source(id="3", type="event")
    context = Context()
    event = Event(id="event-id", type="type", source=source, context=context, profile=profile, session=session)
    flow = Flow(id="flow-id", name="flow")
    dot = DotAccessor(profile, session, payload, event, flow)

    assert 'test.id' == dot['test.id']


from tracardi_plugin_sdk.domain.register import Plugin, Spec, MetaData
from tracardi_plugin_sdk.domain.result import Result
from tracardi_plugin_sdk.action_runner import ActionRunner


class StartAction(ActionRunner):

    def __init__(self, *args, **kwargs):
        pass

    async def run(self, void):
        if self.debug and self.profile.id == '@debug-profile-id':
            raise ValueError("Start action can not run in debug mode without connection to Debug action.")
        return Result(port="payload", value={})


def register() -> Plugin:
    return Plugin(
        start=True,
        debug=False,
        spec=Spec(
            module='tracardi.process_engine.action.v1.start_action',
            className='StartAction',
            inputs=["void"],
            outputs=["payload"],
            init=None,
            version='0.1',
            license="MIT",
            author="Risto Kowaczewski"
        ),
        metadata=MetaData(
            name='Start',
            desc='Starts workflow and returns empty payload.',
            type='flowNode',
            keywords=['start node'],
            width=200,
            height=100,
            icon='start',
            group=["Input/Output"]
        )
    )



import pytz
from _datetime import datetime, timedelta
from typing import Optional, Tuple
from pydantic import BaseModel
from enum import Enum


class DatetimeType(str, Enum):
    second = 'second'
    minute = 'minute'
    hour = 'hour'
    day = 'day'
    week = 'week'
    month = 'month'
    year = 'year'


class DateDeltaPayload(BaseModel):
    value: int
    entity: DatetimeType

    def get_delta(self):
        entity = self.entity
        value = abs(self.value)

        if entity in ['second', 'seconds']:
            return timedelta(seconds=value)
        elif entity in ['minute', 'minutes']:
            return timedelta(minutes=value)
        elif entity in ['hour', 'hours']:
            return timedelta(hours=value)
        elif entity in ['day', 'days']:
            return timedelta(days=value)
        elif entity in ['week', 'weeks']:
            return timedelta(weeks=value)
        elif entity in ['month', 'months']:
            return timedelta(days=value * 31)
        elif entity in ['year', 'years']:
            return timedelta(days=value * 356)

        return None


class DatetimePayload(BaseModel):
    second: int = None
    minute: int = None
    hour: int = None
    date: int = None
    month: int = None
    year: int = None
    meridiem: str = None
    timeZone: int = 0

    @staticmethod
    def now():
        now = datetime.utcnow()
        return DatetimePayload(year=now.year, month=now.month, date=now.day,
                               hour=now.hour, minute=now.minute, second=now.second,
                               meridiem=now.strftime("%p"))

    def is_set(self):
        return self.year is not None \
               and self.month is not None \
               and self.date is not None \
               and self.hour is not None \
               and self.minute is not None \
               and self.second is not None \
               and self.meridiem is not None

    def get_date(self) -> [datetime, None]:
        if self.is_set():
            return datetime(year=self.year,
                            month=self.month,
                            day=self.date,
                            hour=self.hour,
                            minute=self.minute,
                            second=self.second)
        return None

    def __str__(self):
        return "{}/{}/{} {}:{}:{} {}{}{:02d}".format(
            self.year,
            self.month,
            self.date,
            self.hour,
            self.minute,
            self.second,
            self.meridiem,
            "+" if self.timeZone >= 0 else "-",
            self.timeZone
        )


class DatePayload(BaseModel):
    delta: Optional[DateDeltaPayload] = None
    absolute: Optional[DatetimePayload] = None

    def get_date(self) -> datetime:
        if self.absolute is None:
            absolute_date = datetime.now()
        else:
            absolute_date = self.absolute.get_date()

            # If absolute date is None, Then use now

            if absolute_date is None:
                absolute_date = datetime.now()

        # Get delta
        if self._is_delta_set():
            delta, sign = self._get_delta()
            return absolute_date + (sign * delta)

        return absolute_date

    def is_absolute(self):
        return self.absolute is not None and not self._is_delta_set()

    def _is_delta_set(self) -> bool:
        return self.delta is not None

    def _get_delta(self) -> (delta, int):
        return self.delta.get_delta(), -1 if self.delta.value < 0 else 1


class DatetimeRangePayload(BaseModel):
    minDate: Optional[DatePayload] = None
    maxDate: Optional[DatePayload] = None
    where: Optional[str] = ""
    timeZone: Optional[str] = None
    offset: int = 0
    limit: int = 20
    rand: Optional[float] = 0

    def get_dates(self) -> (datetime, datetime):

        if self._is_now(self.minDate):
            self.minDate = DatePayload(absolute=DatetimePayload.now())

        if self._is_now(self.maxDate):
            self.maxDate = DatePayload(absolute=DatetimePayload.now())

        print("_is_min_date_absolute", self._is_min_date_absolute())
        print("_is_max_date_absolute", self._is_max_date_absolute())

        # Set Anchor date
        if self._is_min_date_absolute() and not self._is_max_date_absolute():
            self.maxDate.absolute = self.minDate.absolute
        elif not self._is_min_date_absolute() and self._is_max_date_absolute():
            self.minDate.absolute = self.maxDate.absolute
        elif not self._is_min_date_absolute() and not self._is_max_date_absolute():
            self.minDate.absolute = DatetimePayload.now()
            self.maxDate.absolute = DatetimePayload.now()

        min_date = self.minDate.get_date()
        max_date = self.maxDate.get_date()

        if min_date > max_date or min_date == max_date:
            raise ValueError(
                "Incorrect time range. From date `{}` is earlier then to date `{}` or dates are equal.".format(
                    min_date, max_date
                ))

        return min_date, max_date

    def _is_now(self, date: DatePayload):
        return date is None or (date.absolute is None and date.delta is None)

    def _is_min_date_absolute(self):
        return self.minDate is None or self.minDate.is_absolute()

    def _is_max_date_absolute(self):
        return self.maxDate is None or self.maxDate.is_absolute()

    @staticmethod
    def convert_to_local_datetime(utc_datetime, timezone) -> Tuple[datetime, Optional[str]]:
        try:
            local_tz = pytz.timezone(timezone)
            local_dt = utc_datetime.replace(tzinfo=pytz.utc).astimezone(local_tz)
            return local_tz.normalize(local_dt), timezone  # .normalize might be unnecessary
        except pytz.exceptions.UnknownTimeZoneError as e:
            # todo log error
            return utc_datetime, None



from tracardi_graph_runner.domain.flow_graph_data import FlowGraphData, Node, Edge
import tracardi.domain.flow as f
from tracardi.domain.flow import FlowRecord


def test_flow_encoding():
    flow = f.Flow(
        id="1",
        enabled=True,
        name="Flow",
        flowGraph=FlowGraphData(
            nodes=[
                Node(**{
                    "id": "start",
                    "type": "flowNode",
                    "position": {
                        "x": 380,
                        "y": -260
                    },
                    "data": {
                        "start": True,
                        "metadata": {
                            "name": "Start",
                            "type": "flowNode",
                            "desc": "Get payload",
                            "width": 100,
                            "height": 100,
                            "icon": "payload"
                        },
                        "spec": {
                            "className": "PayloadPlugin",
                            "module": "tracardi.tests.manual.plugins.payload.payload_plugin",
                            "inputs": [],
                            "outputs": [
                                "payload"
                            ],
                            'init': {"ala-ma-kota": "a-kot"}
                        }
                    }
                }),
                Node(**{
                    "id": "763cbc92-5236-4d30-bff2-a10dbbf8c83f",
                    "type": "flowNode",
                    "position": {
                        "x": 380,
                        "y": -60
                    },
                    "data": {
                        "metadata": {
                            "name": "Log",
                            "type": "flowNode",
                            "width": 200,
                            "height": 100,
                            "icon": "debug"
                        },
                        "spec": {
                            "className": "DebugPlugin",
                            "module": "tracardi.tests.manual.plugins.debug.debug_plugin",
                            "inputs": [
                                "payload"
                            ],
                            "outputs": []
                        }
                    }
                })
            ],
            edges=[
                Edge(**
                     {
                         "source": "start",
                         "sourceHandle": "payload",
                         "target": "763cbc92-5236-4d30-bff2-a10dbbf8c83f",
                         "targetHandle": "payload",
                         "id": "reactflow__edge-9f95ad82-9d7c-4f61-a439-580ba4c730e3payload-763cbc92-5236-4d30-bff2-a10dbbf8c83fpayload",
                         "type": "default"
                     })
            ]
        )
    )

    record = FlowRecord.encode(flow)
    new_flow = record.decode()
    assert new_flow.flowGraph.nodes[0].data.spec.init == {"ala-ma-kota": "a-kot"}






from pydantic import BaseModel
from tracardi.domain.enum.yes_no import YesNo


class Settings(BaseModel):
    enabled: bool = True
    hidden: bool = False

    @staticmethod
    def as_bool(state: YesNo):
        return state.value == state.yes



from device_detector import DeviceDetector
from tracardi_plugin_sdk.domain.register import Plugin, Spec, MetaData
from tracardi_plugin_sdk.domain.result import Result
from tracardi_plugin_sdk.action_runner import ActionRunner


class DetectClientAgentAction(ActionRunner):

    def __init__(self, *args, **kwargs):
        pass

    async def run(self, void):

        try:
            if not isinstance(self.session.context, dict):
                raise KeyError("No session context defined.")

            ua = self.session.context['browser']['browser']['userAgent']

            detector = DeviceDetector(ua, skip_bot_detection=False)
            device = detector.parse()

            response = {
                'status': {
                    "code": 200,
                    "message": "OK"
                },
                "device": {
                    "model": {
                        "name": device.device_model(),
                        "brand": {
                            "name": device.device_brand_name(),
                        },
                        "type": device.device_type()
                    },
                    "os": {
                        "name": device.os_name(),
                        "version": device.os_version(),
                    },
                    "client": {
                        "type": device.client_type(),
                        "name": device.client_name(),
                        "version": device.secondary_client_version(),
                        "engine": device.engine(),
                    },
                    "type": {
                        "mobile": device.is_mobile(),
                        "desktop": device.is_desktop(),
                        "bot": device.is_bot(),
                        "tv": device.is_television(),
                    }

                }
            }
        except KeyError as e:
            response = {
                'status': {
                    "code": 500,
                    "message": str(e)
                },
                "device": {
                    "model": {
                        "name": 'Unknown',
                        "brand": {
                            "name": 'Unknown',
                        },
                        "type": 'Unknown'
                    },
                    "os": {
                        "name": 'Unknown',
                        "version": 'Unknown',
                    },
                    "client": {
                        "type": 'Unknown',
                        "name": 'Unknown',
                        "version": 'Unknown',
                        "engine": 'Unknown',
                    },
                    "type": {
                        "mobile": 'Unknown',
                        "desktop": 'Unknown',
                        "bot": 'Unknown',
                        "tv": 'Unknown',
                    }
                }
            }

        return Result(port="client-info", value=response)


def register() -> Plugin:
    return Plugin(
        start=False,
        debug=False,
        spec=Spec(
            module='tracardi.process_engine.action.v1.detect_client_agent_action',
            className='DetectClientAgentAction',
            inputs=["void"],
            outputs=["client-info"],
            init=None,
            manual="detect_client_agent_action",
            version='0.1',
            license="MIT",
            author="Risto Kowaczewski"
        ),
        metadata=MetaData(
            name='Get client agent',
            desc='It will parse any user agent and detect the browser, operating system, device used (desktop, '
                 'tablet, mobile, tv, cars, console, etc.), brand and model. It detects thousands '
                 'of user agent strings, even from rare and obscure browsers and devices. It returns an object containing '
                 'all the information',
            type='flowNode',
            width=100,
            height=100,
            icon='browser',
            group=["Processing"]
        )
    )



from pprint import pprint
from uuid import uuid4

from tracardi.domain.context import Context
from tracardi.domain.event import Event
from tracardi.domain.events import Events
from tracardi.domain.flow import FlowRecord
from tracardi.domain.named_entity import NamedEntity
from tracardi.domain.profile import Profile
from tracardi.domain.rule import Rule
from tracardi.domain.session import Session
from tracardi.domain.type import Type
from tracardi.process_engine.rules_engine import RulesEngine
import asyncio


if __name__ == "__main__":
    async def main():
        source = NamedEntity(id="mobile-app", name="mobile-app")

        # Create profile, rule needs it
        p = Profile.new()
        await p.storage().save()
        profile = await p.storage().load()
        assert p.id == profile.id
        assert isinstance(profile, Profile)

        flow_record = FlowRecord(id="1", name="flow-1")
        await flow_record.storage().save()

        rule = Rule(
            id="string",
            source=source,
            name="my-rule",
            event=Type(type="xxx1"),
            flow=flow_record
        )
        await rule.storage().save()

        saved_rule = await rule.storage().load()
        assert saved_rule.id == rule.id
        x = await RulesEngine.load_rules("xxx2")
        pprint(x)

        payload = {
            "id": str(uuid4()),
            "event_server": Context(),
            "source": source.dict(),
            "profile": p.dict(),
            "context": {},
            "session": {"id": "0e6121a5-3ea0-45ed-a9ad-552e1765167f",
                        "event_server": {
                            "page": {"url": "http://localhost:8002/tracker/", "path": "/tracker/", "hash": "",
                                     "title": "My title",
                                     "referer": {"host": None, "query": None}, "history": {"length": 2}}},
                        },
            "properties": {"a": "tak"}, "type": "xxx1", "user": {"id": "user-id-2"}
        }

        event = Event(**payload)
        events = Events()
        events.append(event)
        events.append(event)
        session = Session(id="session-id")
        profile = Profile(id="profile-id")

        rules_engine = RulesEngine(
            session,
            profile,
            events)

        flow_result, segmentation_result, console_log = await rules_engine.execute(source.id)
        # print(stats.to_json())
        print(flow_result, segmentation_result)

        await rule.storage().delete()


    asyncio.run(main())



from typing import Optional, List

from pydantic import BaseModel

from tracardi.domain.entity import Entity
from tracardi.service.storage.crud import StorageCrud


class ResponseItem(BaseModel):
    index: str
    slices: Optional[List[str]] = None


class Response(Entity):
    data: List[ResponseItem]

    def storage(self) -> StorageCrud:
        return StorageCrud("response", Response, entity=self)



from datetime import datetime
from typing import Optional, Any

from .entity import Entity
from .metadata import Metadata
from .named_entity import NamedEntity
from .time import Time
from tracardi.service.storage.crud import StorageCrud
from .type import Type


class Rule(Entity):
    name: str
    event: Type
    flow: NamedEntity

    source: Optional[NamedEntity] = None
    enabled: Optional[bool] = True
    description: Optional[str] = "No description provided"
    properties: Optional[dict] = None
    metadata: Optional[Metadata]

    def __init__(self, **data: Any):
        data['metadata'] = Metadata(
            time=Time(
                insert=datetime.utcnow()
            ))
        super().__init__(**data)

    def storage(self) -> StorageCrud:
        return StorageCrud("rule", Rule, entity=self)



from uuid import uuid4
from elasticsearch import helpers, AsyncElasticsearch
from elasticsearch._async.client.sql import SqlClient
from elasticsearch._async.client.utils import SKIP_IN_PATH, query_params
from elasticsearch.exceptions import NotFoundError

from ssl import create_default_context

from tracardi import config
from tracardi.domain.value_object.bulk_insert_result import BulkInsertResult

_singleton = None


class AwsCompatibleSqlClient(SqlClient):

    @query_params()
    async def clear_cursor(self, body, params=None, headers=None):
        """
        Clears the SQL cursor

        `<https://www.elastic.co/guide/en/elasticsearch/reference/7.12/sql-pagination.html>`_

        :arg body: Specify the cursor value in the `cursor` element to
            clean the cursor.
        """
        if body in SKIP_IN_PATH:
            raise ValueError("Empty value passed for a required argument 'body'.")

        return await self.transport.perform_request(
            "POST", "/_sql/close", params=params, headers=headers, body=body
        )

    @query_params("format")
    async def query(self, body, params=None, headers=None):
        """
        Executes a SQL request

        `<https://www.elastic.co/guide/en/elasticsearch/reference/7.12/sql-rest-overview.html>`_

        :arg body: Use the `query` element to start a query. Use the
            `cursor` element to continue a query.
        :arg format: a short version of the Accept header, e.g. json,
            yaml
        """
        if body in SKIP_IN_PATH:
            raise ValueError("Empty value passed for a required argument 'body'.")

        return await self.transport.perform_request(
            "POST", "/_sql", params=params, headers=headers, body=body
        )

    @query_params()
    async def translate(self, body, params=None, headers=None):

        if body in SKIP_IN_PATH:
            raise ValueError("Empty value passed for a required argument 'body'.")

        return await self.transport.perform_request(
            config.elastic.sql_translate_method, config.elastic.sql_translate_url, params=params, headers=headers,
            body=body
        )


class Elastic:

    def __init__(self, **kwargs):
        self._cache = {}
        self._client = AsyncElasticsearch(**kwargs)
        self._client.sql = AwsCompatibleSqlClient(self._client)

    async def translate(self, sql):
        return await self._client.sql.translate(body=sql)

    async def close(self):
        await self._client.close()

    async def get(self, index, id):
        return await self._client.get(index=index, doc_type='_doc', id=id)

    # todo error handling move to service
    async def delete(self, index, id):
        try:
            return await self._client.delete(index=index, doc_type="_doc", id=id)
        except NotFoundError:
            return None

    async def delete_by_query(self, index, body):
        try:
            return await self._client.delete_by_query(index=index, body=body)
        except NotFoundError:
            return None

    # todo error handling move to service
    async def exists(self, index, id):
        try:
            return await self._client.exists(index=index, doc_type="_doc", id=id)
        except NotFoundError:
            return None

    async def search(self, index, query):
        return await self._client.search(index=index, body=query)

    async def scan(self, index, query):

        _generator = helpers.async_scan(
            self._client,
            query=query,
            index=index,
        )

        async for doc in _generator:
            yield doc

    async def insert(self, index, records) -> BulkInsertResult:

        if not isinstance(records, list):
            raise ValueError("Insert expects payload to be list.")

        bulk = []
        ids = []
        for record in records:

            if '_id' in record:
                _id = record['_id']
                del (record['_id'])
            else:
                _id = str(uuid4())

            ids.append(_id)
            record = {
                "_index": index,
                "_id": _id,
                "_source": record
            }

            bulk.append(record)

        success, errors = await helpers.async_bulk(self._client, bulk)

        return BulkInsertResult(
            saved=success,
            errors=errors,
            ids=ids
        )

    async def update(self, index, id, record):
        return await self._client.update(index, body=record, id=id)

    async def remove_index(self, index):
        return await self._client.indices.delete(index=index)

    async def create_index(self, index, mappings):
        return await self._client.indices.create(index=index, ignore=400, body=mappings)

    async def exists_index(self, index):
        return await self._client.indices.exists(index=index)

    async def list_indices(self):
        return await self._client.indices.get("*")

    async def refresh(self, index, params=None, headers=None):
        return await self._client.indices.refresh(index=index, params=params, headers=headers)

    async def flush(self, index, params=None, headers=None):
        return await self._client.indices.flush(index=index, params=params, headers=headers)

    @staticmethod
    def _get_elastic_config():

        kwargs = {}

        if config.elastic.host:
            kwargs['hosts'] = config.elastic.host
        if config.elastic.scheme:
            kwargs['scheme'] = config.elastic.scheme
        if config.elastic.sniffer_timeout:
            kwargs['sniffer_timeout'] = config.elastic.sniffer_timeout
        if config.elastic.sniff_on_start:
            kwargs['sniff_on_start'] = config.elastic.sniff_on_start
        if config.elastic.sniff_on_connection_fail:
            kwargs['sniff_on_connection_fail'] = config.elastic.sniff_on_connection_fail
        if config.elastic.maxsize:
            kwargs['maxsize'] = config.elastic.maxsize

        if config.elastic.cafile:
            context = create_default_context(cafile=config.elastic.cafile)
            kwargs['ssl_context'] = context

        if config.elastic.http_auth_password and config.elastic.http_auth_username:
            kwargs['http_auth'] = (config.elastic.http_auth_username, config.elastic.http_auth_password)

        if config.elastic.cloud_id:
            kwargs['cloud_id'] = config.elastic.cloud_id

        if config.elastic.api_key:
            kwargs['api_key'] = config.elastic.api_key

        if config.elastic.http_compress:
            kwargs['http_compress'] = config.elastic.http_compress

        return kwargs

    @staticmethod
    def instance():

        global _singleton

        def get_elastic_client():
            kwargs = Elastic._get_elastic_config()
            return Elastic(**kwargs)

        if _singleton is None:
            _singleton = get_elastic_client()

        return _singleton



from datetime import datetime
from typing import Optional, Any
from uuid import uuid4

from .entity import Entity
from .metadata import Metadata
from .time import Time
from tracardi.service.storage.crud import StorageCrud
from .value_object.operation import Operation


class Session(Entity):
    metadata: Optional[Metadata]
    operation: Operation = Operation()
    profile: Optional[Entity] = None
    context: Optional[dict] = {}
    properties: Optional[dict] = {}

    def __init__(self, **data: Any):
        data['metadata'] = Metadata(
            time=Time(
                insert=datetime.utcnow()
            ))
        super().__init__(**data)

    def replace(self, session):
        self.metadata = session.metadata
        self.profile = session.profile
        self.context = session.context
        self.id = session.id
        self.properties = session.properties
        self.operation = session.operation

    def storage(self, **kwargs) -> StorageCrud:
        return StorageCrud("session", Session, entity=self)

    @staticmethod
    def new() -> 'Session':
        return Session(id=str(uuid4()))



from typing import Optional, Any
from pydantic import BaseModel


class PII(BaseModel):

    """
    Personally identifiable information, or PII, is any data that could
    potentially be used to identify a particular person. Examples include a full name,
    Social Security number, driver's license number, bank account number,
    passport number, and email address.
    """

    name: Optional[Any] = None
    surname: Optional[Any] = None
    birthDate: Optional[Any] = None
    email: Optional[Any] = None
    telephone: Optional[Any] = None
    other: Optional[dict] = {}




import logging
from tracardi.event_server.service.persistence_service import PersistenceService
from tracardi.service.storage.collection_crud import CollectionCrud
from tracardi.service.storage.elastic_storage import ElasticStorage


_logger = logging.getLogger('FlowActionPlugins')


class FlowActionPlugins(list):

    def storage(self) -> PersistenceService:
        return PersistenceService(ElasticStorage(index_key="action"))

    def bulk(self) -> CollectionCrud:
        return CollectionCrud("action", self)



def replace_with_ellipsis(dictionary, key, value=...):
    return {k: replace_with_ellipsis(value if v == key else v, key, value)
            for k,v in dictionary.items()} if isinstance(dictionary, dict) else dictionary



class SegmentStats:

    def __init__(self, triggered=0, errors=None, segments=None):
        if segments is None:
            segments = []
        self.segments = segments
        if errors is None:
            errors = []
        self.triggered = triggered
        self.errors = errors

    def __add__(self, other):
        if not isinstance(other, SegmentStats):
            raise ValueError("SegmentStats can add only other SegmentStats. {} given".format(type(other)))

        self.triggered += other.triggered
        self.errors += other.errors
        self.segments += other.segments

        return self

    def to_json(self):
        return {
            "triggered": self.triggered,
            'errors': self.errors,
            'segments': self.segments
        }



from fastapi import APIRouter
from fastapi import HTTPException

from tracardi.config import memory_cache
from tracardi.domain.entity import Entity
from tracardi.domain.source import Source, SourceRecord
from tracardi.event_server.utils.memory_cache import MemoryCache, CacheItem


router = APIRouter()


class SourceCacher:

    def __init__(self):
        self._cache = MemoryCache()

    async def validate_source(self, source_id) -> Source:
        entity = Entity(id=source_id)

        source = await self.get(entity)  # type: Source
        if source is None:
            raise HTTPException(detail="Access denied. Invalid source.", status_code=401)

        if not source.enabled:
            raise HTTPException(detail="Access denied. Source disabled.", status_code=404)

        return source

    async def get(self, source: Entity):
        if 'source' in self._cache:
            source = self._cache['source'].data
            return source
        else:
            # Expired
            source = await source.storage("source").load(SourceRecord)  # type: SourceRecord
            if source is not None:
                self._cache['source'] = CacheItem(data=source, ttl=memory_cache.source_ttl)
                return source
            return None


source_cache = SourceCacher()



import os
from lark import Lark

_local_dir = os.path.dirname(__file__)


class Parser:

    def __init__(self, grammar, start, parser='earley', transformer=None):
        import_paths = [
            os.path.join(_local_dir, 'grammar')
        ]
        self.base_parser = Lark(grammar,
                                start=start,
                                transformer=transformer,
                                import_paths=import_paths)

    @staticmethod
    def read(file):
        with open(os.path.join(_local_dir, file)) as f:
            return f.read()

    def parse(self, query):
        return self.base_parser.parse(query)



from typing import Optional
from pydantic import BaseModel

from ..context import Context
from ..entity import Entity
from ..event import Event


class EventPayload(BaseModel):
    type: str
    properties: Optional[dict] = {}
    user: Optional[Entity] = None
    options: Optional[dict] = {}

    def is_persistent(self):
        if 'save' in self.options and isinstance(self.options['save'], bool):
            return self.options['save']
        else:
            return True

    def to_event(self, metadata, source, session, profile, options):
        return Event.new({
            "metadata": metadata,
            "session": session,
            "profile": profile.dict(exclude={"operation": ...}),
            "user": self.user,
            "type": self.type,
            "properties": self.properties,
            "source": source,  # Entity
            'context': Context(config=options, params={})
        })




from typing import Optional
from pydantic import BaseModel
from tracardi.domain.pii import PII


class Private(BaseModel):
    pii: Optional[PII] = None


class ProfileTraits(BaseModel):
    private: Optional[dict] = {}
    public: Optional[dict] = {}




from dotty_dict import dotty


class DotNotationConverter:

    def __init__(self, profile):
        self.profile = profile

    def get_profile_fiel_value_pair(self, dot_notation):
        if dot_notation.startswith('profile@'):
            field = dot_notation[len('profile@'):]
            dot = dotty(self.profile.dict())
            try:
                return field, dot[field]
            except KeyError:
                raise KeyError("Invalid dot notation. Could not find value for `{}` in profile.".format(dot_notation))
        else:
            raise KeyError(
                "Invalid dot notation. Only values from profile are allowed as merge_keys. `{}` does not start with profile@..".format(
                    dot_notation))


from collections import defaultdict
from time import time
from typing import Dict, List, Tuple

from pydantic import ValidationError
from tracardi_graph_runner.domain.debug_info import DebugInfo
from tracardi_graph_runner.domain.error_debug_info import ErrorDebugInfo
from tracardi_graph_runner.domain.debug_info import FlowDebugInfo
from tracardi_graph_runner.domain.flow_history import FlowHistory
from tracardi_graph_runner.domain.work_flow import WorkFlow
from tracardi_plugin_sdk.domain.console import Log
from ..domain.console import Console, ConsoleLog
from ..domain.entity import Entity
from ..domain.flow import Flow
from ..domain.profile import Profile
from ..domain.record.event_debug_record import EventDebugRecord
from ..domain.session import Session
from ..event_server.utils.memory_cache import MemoryCache, CacheItem
from ..domain.event import Event
from ..domain.events import Events
from ..domain.rule import Rule
from ..domain.rules import Rules
from ..service.storage.collection_crud import CollectionCrud
import asyncio


class RulesEngine:

    def __init__(self,
                 session: Session,
                 profile: Profile,
                 events: Events
                 ):
        self.session = session
        self.profile = profile
        self.events = events

    @staticmethod
    async def load_rules(event_type: str):

        query = {
            "query": {
                "bool": {
                    "must": [
                        {
                            "term": {
                                "event.type.keyword": event_type
                            }
                        },
                        {
                            "match": {
                                "enabled": True
                            }
                        }
                    ]
                }
            }
        }

        # todo set MemoryCache ttl from env
        memory_cache = MemoryCache()

        if 'rules' not in memory_cache:
            flows_data = await Rules.storage().filter(query)
            memory_cache['rules'] = CacheItem(data=flows_data, ttl=1)

        rules = list(memory_cache['rules'].data)
        return rules

    async def invoke(self, source_id=None) -> Tuple[Dict[str, List[Dict[str, DebugInfo]]], Dict[str, list], ConsoleLog]:

        flow_task_store = defaultdict(list)
        debug_info_by_event_type_and_rule_name = defaultdict(list)

        for event in self.events:  # type: Event

            # Loads rules only for event.type
            rules = await RulesEngine.load_rules(event.type)

            for rule in rules:

                try:
                    rule = Rule(**rule)
                except ValidationError as e:
                    print(str(e))
                    # todo log error and disable rule
                    continue

                if not rule.enabled:
                    continue

                try:
                    flow = await Flow.decode(rule.flow.id)
                except Exception as e:
                    # This is empty DebugInfo without nodes
                    debug_info = DebugInfo(
                        timestamp=time(),
                        flow=FlowDebugInfo(
                            id=rule.flow.id,
                            error=[ErrorDebugInfo(msg=str(e), file=__file__, line=103)]
                        ),
                        event=Entity(id=event.id)
                    )
                    debug_info_by_event_type_and_rule_name[event.type].append({rule.name: debug_info})
                    continue

                if not flow.enabled:
                    continue

                # Validates rule source. Type was verified before

                if source_id:
                    if source_id == event.source.id:
                        # todo FlowHistory is empty
                        workflow = WorkFlow(
                            FlowHistory(history=[]),
                            self.session,
                            self.profile,
                            event
                        )
                        flow_task = asyncio.create_task(workflow.invoke(flow, debug=False))
                        flow_task_store[event.type].append((rule.flow.id, event.id, rule.name, flow_task))
                else:
                    # todo FlowHistory is empty
                    workflow = WorkFlow(
                        FlowHistory(history=[]),
                        self.session,
                        self.profile,
                        event
                    )
                    flow_task = asyncio.create_task(workflow.invoke(flow, debug=False))
                    flow_task_store[event.type].append((rule.flow.id, event.id, rule.name, flow_task))

        # Run and report async
        console_log = ConsoleLog()
        for event_type, tasks in flow_task_store.items():
            for flow_id, event_id, rule_name, task in tasks:
                try:
                    debug_info, log_list = await task  # type: DebugInfo, List[Log]

                    # Store logs in one console log
                    for log in log_list:  # type: Log
                        console = Console(
                            origin="node",
                            event_id=event_id,
                            flow_id=flow_id,
                            module=log.module,
                            class_name=log.class_name,
                            type=log.type,
                            message=log.message
                        )
                        console_log.append(console)

                except Exception as e:
                    # todo log error

                    console = Console(
                        origin="workflow",
                        event_id=event_id,
                        flow_id=flow_id,
                        module='tracardi.process_engine.rules_engine',
                        class_name="RulesEngine",
                        type="error",
                        message=str(e)
                    )
                    console_log.append(console)

                    debug_info = DebugInfo(
                        timestamp=time(),
                        flow=FlowDebugInfo(
                            id=flow_id,
                            error=[ErrorDebugInfo(msg=str(e), file=__file__, line=86)]
                        ),
                        event=Entity(id=event_id)
                    )

                debug_info_by_event_type_and_rule_name[event_type].append({rule_name: debug_info})

        # Save profile

        segmentation_info = {"errors": [], "ids": []}
        try:
            save_tasks = []

            # Segmentation
            if self.profile.operation.needs_update() or self.profile.operation.needs_segmentation():
                # Segmentation runs only if profile was updated or flow forced it
                async for event_type, segment_id, error in self.profile.segment(event_types=flow_task_store.keys()):
                    # Segmentation triggered
                    if error:
                        segmentation_info['errors'].append(error)
                    segmentation_info['ids'].append(segment_id)

            # Merging, schedule save only if there is an update in flow.

            if self.profile.operation.needs_merging() and self.profile.operation.needs_update():

                disabled_profiles = await self.profile.merge(limit=2000)
                if disabled_profiles is not None:
                    save_old_profiles_task = asyncio.create_task(disabled_profiles.bulk().save())
                    save_tasks.append(save_old_profiles_task)

            # Must be the last operation
            if self.profile.operation.needs_update():
                save_tasks.append(asyncio.create_task(self.profile.storage().save()))

            # run save tasks
            await asyncio.gather(*save_tasks)

        except Exception as e:
            # this error is a global segmentation error
            # todo log it.
            pass

        finally:
            return debug_info_by_event_type_and_rule_name, segmentation_info, console_log

    @staticmethod
    def _mark_profiles_as_merged(profiles, merge_with) -> List[Profile]:
        disabled_profiles = []

        for profile in profiles:
            profile.active = False
            profile.mergedWith = merge_with
            disabled_profiles.append(profile)

        return disabled_profiles

    def _get_merging_keys_and_values(self):
        merge_key_values = self.profile.get_merge_key_values()

        # Add keyword
        merge_key_values = [(f"{field}.keyword", value) for field, value in merge_key_values if value is not None]

        return merge_key_values

    async def execute(self, source_id) -> Tuple[Dict[str, List[Dict[str, DebugInfo]]], Dict[str, list], ConsoleLog]:

        """ Runs rules and flow and saves debug info and console log """

        debug_info_by_event_type_and_rule_name, segmentation_info, console_log = await self.invoke(source_id)

        # Save console log

        if console_log:
            await console_log.bulk().save()

        # Collect debug info

        record = []
        for debug_info_record in EventDebugRecord.encode(
                debug_info_by_event_type_and_rule_name):  # type: EventDebugRecord
            record.append(debug_info_record)

        # Save in debug index
        bulk = CollectionCrud("debug-info", record)
        await bulk.save()

        return debug_info_by_event_type_and_rule_name, segmentation_info, console_log




from tracardi_plugin_sdk.domain.register import Plugin, Spec, MetaData
from tracardi_plugin_sdk.action_runner import ActionRunner
from tracardi_plugin_sdk.domain.result import Result


class SegmentProfileAction(ActionRunner):

    def __init__(self, *args, **kwargs):
        pass

    async def run(self, void):
        self.profile.operation.segment = True
        return Result(value={}, port="void")


def register() -> Plugin:
    return Plugin(
        start=False,
        spec=Spec(
            module='tracardi.process_engine.action.v1.operations.segment_profile_action',
            className='SegmentProfileAction',
            inputs=["void"],
            outputs=["void"],
            init=None,
            manual="segment_profiles_action"
        ),
        metadata=MetaData(
            name='Segment profile',
            desc='Segment profile when flow ends.This action forces segmentation on profile after flow ends. See '
                 'documentation for more information.',
            type='flowNode',
            width=200,
            height=100,
            icon='segment',
            group=["Operations"]
        )
    )



from datetime import datetime


def to_time_range_sql_query(index, time, min: datetime, max: datetime, query=""):
    sql = """SELECT * FROM "{}" WHERE {} BETWEEN '{}' AND '{}'""".format(
        index, time, min.isoformat(),
        max.isoformat())

    if query:
        query = query.replace('"', "'")
        sql += " AND {}".format(query)

    query = {
        "query": sql
    }

    return query


def to_sql_query(index: str, fields: str = "*", query: str = None, limit: int = 20):
    sql = "SELECT {} FROM \"{}\"".format(fields, index)

    if query:
        query = query.replace('"', "'")
        sql += " WHERE {}".format(query)

    sql += " LIMIT {}".format(limit)

    query = {
        "query": sql
    }

    return query



import elasticsearch
from pydantic import BaseModel

from tracardi.event_server.service.persistence_service import PersistenceService
from tracardi.domain.agg_result import AggResult
from tracardi.domain.storage_result import StorageResult
from tracardi.domain.value_object.bulk_insert_result import BulkInsertResult
from tracardi.exceptions.exception import TracardiException, StorageException
from tracardi.service.storage.elastic_storage import ElasticStorage


class CollectionCrud:

    def __init__(self, index, payload):
        self.payload = payload
        self.index = index
        self.storage = PersistenceService(ElasticStorage(index_key=self.index))

    async def save(self) -> BulkInsertResult:
        try:
            if not isinstance(self.payload, list):
                raise TracardiException("CollectionCrud dave payload must be list.")
            data = [p.dict() for p in self.payload if isinstance(p, BaseModel)]
            return await self.storage.upsert(data)

        except elasticsearch.exceptions.ElasticsearchException as e:
            raise StorageException(str(e))

    async def load(self) -> StorageResult:
        try:

            return await self.storage.load_all()

        except elasticsearch.exceptions.ElasticsearchException as e:
            raise StorageException(str(e))

    async def uniq_field_value(self, field) -> AggResult:
        try:
            query = {
                "size": "0",
                "aggs": {
                    "uniq": {
                        "terms": {
                            "field": field
                        }
                    }
                }
            }
            return AggResult('uniq', await self.storage.query(query), return_counts=False)
        except elasticsearch.exceptions.ElasticsearchException as e:
            raise StorageException(str(e))



from tracardi_plugin_sdk.domain.register import Plugin, Spec, MetaData
from tracardi_plugin_sdk.action_runner import ActionRunner


class EndAction(ActionRunner):

    def __init__(self, *args, **kwargs):
        pass

    async def run(self, void):
        return None


def register() -> Plugin:
    return Plugin(
        start=False,
        spec=Spec(
            module='tracardi.process_engine.action.v1.end_action',
            className='EndAction',
            inputs=["void"],
            outputs=[],
            version='0.1',
            license="MIT",
            author="Risto Kowaczewski"

        ),
        metadata=MetaData(
            name='End',
            desc='Ends workflow.',
            type='flowNode',
            width=100,
            height=100,
            icon='stop',
            group=["Input/Output"]
        )
    )



from pydantic import BaseModel
from tracardi_plugin_sdk.action_runner import ActionRunner

from tracardi_plugin_sdk.domain.result import Result


class Say(BaseModel):
    id: int
    name: str


class task1(ActionRunner):
    def __init__(self, condition):
        self.condition = condition

    async def run(self, payload):
        print("Hello 1!")
        print("conditional=", self.condition)
        print("payload", payload)
        if self.condition == 'abc':
            return Result(port="TRUE", value=["1t"]), Result(port="FALSE", value=None)
        else:
            return Result(port="FALSE", value=["1f"]), Result(port="TRUE", value=None)


class task2(ActionRunner):

    async def run(self, NODE2_i):
        print("Hello, 2!")
        print(NODE2_i)
        NODE2_i += [2]
        return Result(port="NODE2_o", value=NODE2_i)


class task3(ActionRunner):

    async def run(self, NODE3_i):
        print("Hello, 3!")
        print(NODE3_i)
        NODE3_i += [3]
        return Result(port="NODE3_o", value=NODE3_i)


class task4(ActionRunner):

    async def run(self, NODE4_2_i=None, NODE4_3_i=None, NODE4_1_i=None):
        print("Hello, 4!")
        print("1={} 2={} 3={}".format(NODE4_1_i, NODE4_2_i, NODE4_3_i))
        if NODE4_2_i:
            NODE4_2_i += [4]
        if NODE4_3_i:
            NODE4_3_i += [4]
        if NODE4_1_i:
            NODE4_1_i += [4]
        return Result(port="NODE4_o", value=(NODE4_2_i, NODE4_3_i, NODE4_1_i))


class task7(ActionRunner):

    async def run(self, payload):
        print("Hello, 2!")
        return Result(port="NODE2_o", value=[2])

class task8(ActionRunner):

    async def run(self, INP1=None, INP2=None, INP3=None):
        print("Hello, 4!")
        print("1={} 2={} 3={}".format(INP1, INP2, INP3))
        if INP1:
            INP1 += [4]
        if INP2:
            INP2 += [4]
        if INP3:
            INP3 += [4]
        return Result(port="NODE4_o", value=(INP1, INP2, INP3))


class task5(ActionRunner):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.x = 1

    async def run(self, **kwargs):
        print("Hello, 5!")
        print(kwargs)
        print()
        return Result(port="NODE5_o", value=5)


class task6(ActionRunner):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.x = 1

    async def run(self, **kwargs):
        print("Hello, 6!")
        print(kwargs)
        return Result(port="say6_o", value=6)



from typing import Optional

from pydantic import BaseModel

from tracardi.domain.consent import Consent


class Consents(BaseModel):
    required: Optional[bool] = True
    list: Optional[Consent] = None
    data_controller_contact: Optional[str] = "Contact site owner"



import pytest
from tracardi.domain.entity import Entity
from tracardi.domain.payload.event_payload import EventPayload
from tracardi.domain.payload.tracker_payload import TrackerPayload
from tracardi.domain.profile import Profile
from tracardi.domain.session import Session


@pytest.mark.asyncio
async def test_save_events():
    tracker_payload = TrackerPayload(
        session=Entity(id="12345"),
        profile=Profile.new(),
        source=Entity(id="scope"),
        events=[
            EventPayload(
                type="click",
                properties={"btn": [1, 2, 3]},
                options={"save": True}
            ),
            EventPayload(
                type="click",
                properties={"btn": [3, 4, 5]},
                options={"save": True}
            )
        ]
    )
    events = tracker_payload.get_events(Session.new(), Profile.new())
    result = await events.bulk().save()
    assert result.saved == 2  # profile id must be form session

    for id in result.ids:
        entity = Entity(id=id)
        await entity.storage("event").delete()




from tracardi_plugin_sdk.domain.register import Plugin, Spec, MetaData
from tracardi_plugin_sdk.domain.result import Result
from tracardi_plugin_sdk.action_runner import ActionRunner


class ReadEventAction(ActionRunner):

    def __init__(self, **kwargs):
        pass

    async def run(self, void):
        return Result(port="event", value=self.event.dict())


def register() -> Plugin:
    return Plugin(
        start=False,
        debug=False,
        spec=Spec(
            module='tracardi.process_engine.action.v1.read_event_action',
            className='ReadEventAction',
            inputs=['void'],
            outputs=["event"],
            init=None,
            version='0.1',
            license="MIT",
            author="Risto Kowaczewski"
        ),
        metadata=MetaData(
            name='Read event',
            desc='Loads event into workflow.',
            type='flowNode',
            width=200,
            height=100,
            icon='event',
            group=["Read"]
        )
    )



from tracardi.domain.storage_result import StorageResult
from tracardi.event_server.service.persistence_service import PersistenceService
from tracardi.domain.value_object.bulk_insert_result import BulkInsertResult
from tracardi.service.storage.elastic_storage import ElasticStorage
import tracardi.domain.entity as domain


class BaseStorageCrud:

    def __init__(self, index, entity, exclude=None, exclude_unset=False):
        self.exclude_unset = exclude_unset
        self.exclude = exclude
        self.index = index
        self.entity = entity

    async def delete(self) -> dict:
        service = self._get_storage_service()
        return await service.delete(self.entity.id)

    def _get_storage_service(self):
        return PersistenceService(ElasticStorage(index_key=self.index))


class EntityStorageCrud(BaseStorageCrud):

    async def load(self, domain_class_ref=None):
        service = self._get_storage_service()
        data = await service.load(self.entity.id)

        if data:
            if domain_class_ref is None:
                return domain.Entity(**data)
            return domain_class_ref(**data)

        return None

    async def load_by(self, field, value) -> StorageResult:
        service = self._get_storage_service()
        return await service.load_by(field, value)

    async def delete_by(self, field, value) -> dict:
        service = self._get_storage_service()
        return await service.delete_by(field, value)

    async def save(self, row=None):
        if row is None:
            row = {}
        row['id'] = self.entity.id
        service = self._get_storage_service()
        return await service.upsert(row)


class StorageCrud(BaseStorageCrud):

    def __init__(self, index, domain_class_ref, entity, exclude=None, exclude_unset=False):
        super().__init__(index, entity, exclude, exclude_unset)
        self.domain_class_ref = domain_class_ref

    async def load(self):
        service = self._get_storage_service()
        data = await service.load(self.entity.id)

        if data:
            return self.domain_class_ref(**data)

        return None

    async def save(self) -> BulkInsertResult:
        service = self._get_storage_service()
        return await service.upsert(self.entity.dict(exclude_unset=self.exclude_unset, exclude=self.exclude))



import base64
import json
from typing import Dict, List

from pydantic import BaseModel
from tracardi_graph_runner.domain.debug_info import DebugInfo

from tracardi.domain.entity import Entity
from tracardi.domain.value_object.tracker_payload_result import TrackerPayloadResult
from tracardi.service.secrets import b64_encoder, b64_decoder
from tracardi.service.storage.crud import StorageCrud


class EventDebugRecord(Entity):
    content: str = None

    @staticmethod
    def encode(stat: Dict[str, List[Dict[str, DebugInfo]]]) -> List['EventDebugRecord']:

        for event_type, debugging in stat.items():
            for debug_infos in debugging:
                for rule_id, debug_info in debug_infos.items():  # type: DebugInfo
                    # todo - to pole jest za małe (wyskakuje błąd gdy debug infor ma powyżej 32000 znaków)
                    b64 = b64_encoder(debug_info.dict())
                    yield EventDebugRecord(id=debug_info.event.id, content=b64)

    def decode(self) -> DebugInfo:
        # todo - to pole jest za małe (wyskakuje błąd gdy debug infor ma powyżej 32000 znaków)
        debug_info = b64_decoder(self.content)

        return DebugInfo(
            **debug_info
        )

    # Persistence

    def storage(self) -> StorageCrud:
        return StorageCrud("debug-info", EventDebugRecord, entity=self)



import tracardi
from typing import List
from pydantic import BaseModel

from tracardi.event_server.service.persistence_service import PersistenceService
from tracardi.service.storage.elastic_storage import ElasticStorage


class Operation(BaseModel):
    new: bool = False
    update: bool = False
    segment: bool = False
    merge: List[str] = []

    def needs_segmentation(self):
        return self.segment is True

    def needs_update(self):
        return self.update is True

    def needs_merging(self):
        return isinstance(self.merge, list) and len(self.merge) > 0

    @staticmethod
    async def load_profiles_to_merge(merge_key_values: List[tuple], limit=1000) -> List['tracardi.domain.profile.Profile']:
        profiles = await PersistenceService(ElasticStorage('profile')).load_by_values(merge_key_values, limit)
        return [tracardi.domain.profile.Profile(**profile) for profile in profiles]



import asyncio
from pprint import pprint

from tracardi.domain.payload.tracker_payload import TrackerPayload
from tracardi.service.storage.elastic import Elastic


async def main():
    payload = {
        "metadata": {
            "time": {
                "now": "2021-05-07T15:09:35.298Z",
                "utc": -2
            }
        },
        "source": {"id": "scope"},
        "event_server": {
            "browser": {
                "browser": {
                    "name": "Netscape",
                    "engine": "Gecko",
                    "appVersion": "5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/90.0.4430.72 Safari/537.36",
                    "userAgent": "Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/90.0.4430.72 Safari/537.36",
                    "language": "en-US",
                    "onLine": True,
                    "javaEnabled": False,
                    "cookieEnabled": True
                },
                "device": {
                    "platform": "Linux x86_64"
                }
            },
            "storage": {
                "local": {
                    "ajs_group_id": "null",
                    "ajs_user_id": "null",
                    "__user_id": "\"user-id\"",
                    "ajs_group_properties": "{}",
                    "ajs_user_traits": "{}",
                    "__anon_id": "\"ed964435-ea00-4f5a-86d3-89ca0accdeb5\"",
                    "ajs_anonymous_id": "\"6e6c0ef9-d736-4a03-a8fd-f884698481b7\"",
                    "debug": "undefined",
                    "__user_traits": "{\"a\":1}"
                },
                "cookie": {
                    "cookies1": "tracardi-session-id=9ec4500a-057e-4a05-b460-8f132e73ef63; ajs_group_id=null; ajs_anonymous_id=%226e6c0ef9-d736-4a03-a8fd-f884698481b7%22; ajs_user_id=%22f4ca124298%22; event_server-profile-id=32666cad-657b-494e-a7ef-ec28c349eae2",
                    "cookies2": "tracardi-session-id=9ec4500a-057e-4a05-b460-8f132e73ef63, ajs_group_id=null, ajs_anonymous_id=\"6e6c0ef9-d736-4a03-a8fd-f884698481b7\", ajs_user_id=\"f4ca124298\", event_server-profile-id=32666cad-657b-494e-a7ef-ec28c349eae2"
                }
            },
            "screen": {
                "width": 1920,
                "height": 1080,
                "innerWidth": 1745,
                "innerHeight": 237,
                "availWidth": 1920,
                "availHeight": 1053,
                "colorDepth": 24,
                "pixelDepth": 24
            }
        },
        "profile": {
            "id": "6ec4500a-057e-4a05-b460-8f132e73ef63"
        },
        "session": {
            "id": "6ec4500a-057e-4a05-b460-8f132e73ef63"
        },
        "events": [
            {
                "type": "sessionCreated"
            }
        ]
    }

    context = TrackerPayload(**payload)
    pprint(await context.collect())
    #
    await Elastic.instance().close()


loop = asyncio.get_event_loop()
loop.run_until_complete(main())
loop.close()



from datetime import datetime
from typing import Optional, Any
from uuid import uuid4

from tracardi.service.storage.crud import StorageCrud
from .context import Context
from .entity import Entity
from .metadata import Metadata
from .profile import Profile
from .session import Session
from .time import Time


class Event(Entity):
    metadata: Metadata
    type: str
    properties: Optional[dict] = {}

    source: Entity
    session: Session
    profile: Profile = None
    context: Context

    def __init__(self, **data: Any):
        if 'metadata' in data and isinstance(data['metadata'], Metadata):
            data['metadata'].time = Time(
                insert=datetime.utcnow()
            )
        else:
            data['metadata'] = Metadata(
                time=Time(
                    insert=datetime.utcnow()
                )
            )
        super().__init__(**data)

    def replace(self, event):
        self.id = event.id
        self.type = event.type
        self.properties = event.properties
        self.source = event.source
        self.session = event.session
        self.profile = event.profile
        self.context = event.context

    # Persistence

    def storage(self, **kwargs) -> StorageCrud:
        return StorageCrud("event", Event, entity=self)

    @staticmethod
    def new(data: dict) -> 'Event':
        data['id'] = str(uuid4())
        return Event(**data)



# todo remove after 01-09-2021
# from dotty_dict import dotty
# from tracardi.domain.event import Event
# from tracardi.domain.flow import Flow
# from tracardi.domain.profile import Profile
# from tracardi.domain.session import Session
#
#
# class DotAccessor:
#
#     def __init__(self, profile: Profile, session: Session, payload: dict, event: Event, flow: Flow):
#         self.flow = dotty(flow.dict())
#         self.event = dotty(event.dict())
#         self.payload = dotty(payload)
#         self.session = dotty(session.dict())
#         self.profile = dotty(profile.dict())
#
#     def __delitem__(self, key):
#         if key.startswith('profile@'):
#             key = key[len('profile@'):]
#             del self.profile[key]
#         elif key.startswith('session@'):
#             raise KeyError("Could not set session, session is read only")
#         elif key.startswith('flow@'):
#             raise KeyError("Could not set flow, flow is read only")
#         elif key.startswith('payload@'):
#             key = key[len('payload@'):]
#             del self.payload[key]
#         elif key.startswith('event@'):
#             raise KeyError("Could not delete event, event is read only")
#         else:
#             raise ValueError(
#                 "Invalid dot notation. Accessor not available. " +
#                 "Please start dotted path with one of the accessors: [profile@, session@, payload@, event@] ")
#
#     def __setitem__(self, key, value):
#         if key.startswith('profile@'):
#             key = key[len('profile@'):]
#             self.profile[key] = self.__getitem__(value) if not isinstance(value, dict) else value
#         elif key.startswith('session@'):
#             raise KeyError("Could not set session, session is read only")
#         elif key.startswith('flow@'):
#             raise KeyError("Could not set flow, flow is read only")
#         elif key.startswith('payload@'):
#             key = key[len('payload@'):]
#             self.payload[key] = self.__getitem__(value) if not isinstance(value, dict) else value
#         elif key.startswith('event@'):
#             raise KeyError("Could not set event, event is read only")
#         else:
#             raise ValueError(
#                 "Invalid dot notation. Accessor not available. " +
#                 "Please start dotted path with one of the accessors: [profile@, session@, payload@, event@] ")
#
#     def __getitem__(self, dot_notation):
#         if isinstance(dot_notation, str):
#             if dot_notation.startswith('flow@'):
#                 value = dot_notation[len('flow@'):]
#                 try:
#                     return self.flow[value]
#                 except KeyError:
#                     raise KeyError("Invalid dot notation. Could not find value for `{}` in flow.".format(value))
#             elif dot_notation.startswith('profile@'):
#                 value = dot_notation[len('profile@'):]
#                 try:
#                     return self.profile[value]
#                 except KeyError:
#                     raise KeyError("Invalid dot notation. Could not find value for `{}` in profile.".format(value))
#             elif dot_notation.startswith('session@'):
#                 value = dot_notation[len('session@'):]
#                 try:
#                     return self.session[value]
#                 except KeyError:
#                     raise KeyError("Invalid dot notation. Could not find value for `{}` in session.".format(value))
#             elif dot_notation.startswith('payload@'):
#                 value = dot_notation[len('payload@'):]
#                 try:
#                     return self.payload[value]
#                 except KeyError:
#                     raise KeyError("Invalid dot notation. Could not find value for `{}` in payload.".format(value))
#             elif dot_notation.startswith('event@'):
#                 value = dot_notation[len('event@'):]
#                 try:
#                     return self.event[value]
#                 except KeyError:
#                     raise KeyError("Invalid dot notation. Could not find value for `{}` in event.".format(value))
#         return dot_notation
#
#     def __contains__(self, item):
#         try:
#             self.__getitem__(item)
#             return True
#         except KeyError:
#             return False
#
#     @staticmethod
#     def get(dot_notation, payload, prefix):
#         value = dot_notation[len(prefix + '@'):]
#         try:
#             return payload[value]
#         except KeyError:
#             raise KeyError("Could not find value for `{}` in {}".format(value, prefix))
#
#     @staticmethod
#     def set(key, value, payload, prefix):
#         key = key[len(prefix + '@'):]
#         payload[key] = value



from time import sleep

import pytest
from uuid import uuid4

from tracardi.domain.consent import Consent
from tracardi.domain.entity import Entity
from tracardi.domain.profile import Profile
from tracardi.domain.session import Session


@pytest.mark.asyncio
async def test_domain_persistance():
    async def main(obj):
        obj = obj(id=str(uuid4()))
        result = await obj.storage().load()
        assert result is None
        await obj.storage().save()
        sleep(1)
        object_loaded = await obj.storage().load()
        assert object_loaded.id == obj.id
        await obj.storage().delete()
        sleep(1)
        result = await obj.storage().load()
        assert result is None

    for o in [
        Profile,
        Session,
    ]:
        await main(o)


@pytest.mark.asyncio
async def test_entity_persistance():
    async def main(obj, index):
        entity = Entity(id=str(uuid4()))
        result = await entity.storage(index).load(obj)
        assert result is None
        await entity.storage(index).save()
        sleep(1)
        object_loaded = await entity.storage(index).load(obj)
        assert object_loaded.id == entity.id
        result = await entity.storage(index).load_by("id", entity.id)
        for r in result:
            assert r['id'] == entity.id
        await entity.storage(index).delete()
        sleep(1)
        result = await entity.storage(index).load(obj)
        assert result is None

    for o, index in [
        (Consent, 'consent'),
        (Profile, 'profile'),
        (Session, 'session')
    ]:
        await main(o, index)



from time import sleep

from tracardi.event_server.utils.memory_cache import MemoryCache, CacheItem
from tracardi.exceptions.exception import ExpiredException


def test_memory_cache_add():
    mc = MemoryCache()
    try:
        mc['test'] = 1
        assert False
    except ValueError:
        assert True

    mc['test'] = CacheItem(data={"a": 1}, ttl=10)
    assert mc['test'].data == {"a": 1}


def test_memory_cache_ttl():
    mc = MemoryCache()

    mc['test'] = CacheItem(data={"a": 1}, ttl=2)
    assert 'test' in mc

    sleep(1)
    assert 'test' in mc
    assert mc['test'].data == {"a": 1}

    sleep(2)

    try:
        x = mc['test']
        assert False
    except ExpiredException:
        assert True

    assert 'test' not in mc


def test_memory_cache_usage_scenario():
    mc = MemoryCache()
    if 'x' not in mc:
        mc['x'] = CacheItem(data=1, ttl=2)
    assert mc['x'].data == 1

    sleep(1)

    # must be in cache
    if 'x' not in mc:
        assert False

    assert mc['x'].data == 1

    sleep(2)

    # must not be in cache
    if 'x' in mc:
        assert False




from datetime import datetime
from typing import Any
from pydantic import BaseModel
from tracardi.service.storage.collection_crud import CollectionCrud
from tracardi.service.storage.crud import StorageCrud


class Metadata(BaseModel):
    timestamp: datetime = None

    def __init__(self, **data: Any):
        super().__init__(**data)
        self.timestamp = datetime.utcnow()


class Console(BaseModel):
    metadata: Metadata = Metadata()
    event_id: str
    flow_id: str
    origin: str
    class_name: str
    module: str
    type: str
    message: str

    # Persistence

    def storage(self) -> StorageCrud:
        return StorageCrud("console-log", Console, entity=self)


class ConsoleLog(list):

    def bulk(self) -> CollectionCrud:
        return CollectionCrud("console-log", self)



import subprocess
import importlib
import importlib.util
import sys


def pip_install(package, upgrade=False):
    if upgrade:
        subprocess.check_call([sys.executable, '-m', 'pip', 'install', package, '-U'])
    else:
        subprocess.check_call([sys.executable, '-m', 'pip', 'install', package])


def import_and_install(package, upgrade=False):
    if upgrade:
        pip_install(package.split(".")[0], upgrade=True)

    try:
        return importlib.import_module(package)
    except ImportError:
        pip_install(package.split(".")[0])
    return importlib.import_module(package)


def load_callable(module, className, upgrade=False):
    module = import_and_install(module, upgrade)
    return getattr(module, className)


def is_installed(package_name):
    if package_name in sys.modules or importlib.util.find_spec(package_name) is not None:
        return True
    else:
        return False



from typing import Optional
from pydantic import BaseModel
from tracardi.domain.value_object.save_result import SaveResult


class TrackerPayloadResult(BaseModel):
    session: Optional[SaveResult] = None
    events: SaveResult
    profile: Optional[SaveResult] = None



class FlowStats:

    def __init__(self, triggered=0, errors=None):

        if errors is None:
            errors = []

        self.triggered = triggered
        self.errors = errors

    def __add__(self, other):
        if not isinstance(other, FlowStats):
            raise ValueError("FlowStats can add only other FlowStats. {} given".format(type(other)))

        self.triggered += other.triggered
        self.errors += other.errors

        return self

    def to_json(self):
        return {
                "triggered": self.triggered,
                "errors": self.errors,
        }



def flatten(d):
    if d == {}:
        return d
    else:
        k, v = d.popitem()
        if dict != type(v):
            return {k: v, **flatten(d)}
        else:
            flat_kv = flatten(v)
            for k1 in list(flat_kv.keys()):
                flat_kv[str(k) + '.' + str(k1)] = flat_kv[k1]
                del flat_kv[k1]
            return {**flat_kv, **flatten(d)}



# from typing import Optional, List
# from pydantic import BaseModel
# from tracardi_graph_runner.domain.flow_graph_data import FlowGraphData, Edge, Position, Node
# from tracardi_plugin_sdk.domain.register import MetaData, Plugin, Spec
# from tracardi.service.storage.crud import StorageCrud
# from ..flow import Flow
# from ..named_entity import NamedEntity
# from ...service.secrets import encrypt, decrypt
#
#
# class SpecRecord(BaseModel):
#     className: str
#     module: str
#     inputs: Optional[List[str]] = []
#     outputs: Optional[List[str]] = []
#     init: Optional[str] = ""
#     manual: Optional[str] = None
#
#     @staticmethod
#     def encode(spec: Spec) -> 'SpecRecord':
#         return SpecRecord(
#             className=spec.className,
#             module=spec.module,
#             inputs=spec.inputs,
#             outputs=spec.outputs,
#             init=encrypt(spec.init),
#             manual=spec.manual
#         )
#
#     def decode(self) -> Spec:
#         return Spec(
#             className=self.className,
#             module=self.module,
#             inputs=self.inputs,
#             outputs=self.outputs,
#             init=decrypt(self.init),
#             manual=self.manual
#         )
#
#
# class PluginRecord(BaseModel):
#     start: bool = False
#     debug: bool = False
#     spec: SpecRecord
#     metadata: MetaData
#
#     @staticmethod
#     def encode(plugin: Plugin) -> 'PluginRecord':
#         return PluginRecord(
#             start=plugin.start,
#             debug=plugin.debug,
#             spec=SpecRecord.encode(plugin.spec),
#             metadata=plugin.metadata
#         )
#
#     def decode(self) -> Plugin:
#         data = {
#             "start": self.start,
#             "debug": self.debug,
#             "spec": self.spec.decode(),
#             "metadata": self.metadata
#         }
#         return Plugin.construct(_fields_set=self.__fields_set__, **data)
#
#
# class NodeRecord(BaseModel):
#     id: str
#     type: str
#     position: Position
#     data: PluginRecord
#
#     @staticmethod
#     def encode(node: Node):
#         return NodeRecord(
#             id=node.id,
#             type=node.type,
#             position=node.position,
#             data=PluginRecord.encode(node.data)
#         )
#
#     def decode(self) -> Node:
#         data = {
#             "id": self.id,
#             "type": self.type,
#             "data": self.data.decode(),
#             "position": self.position
#         }
#         return Node.construct(_fields_set=self.__fields_set__, **data)
#
#
# class FlowGraphDataRecord(BaseModel):
#     nodes: List[NodeRecord]
#     edges: List[Edge]
#
#     @staticmethod
#     def encode(flowGraph: FlowGraphData) -> 'FlowGraphDataRecord':
#         return FlowGraphDataRecord(
#             edges=flowGraph.edges,
#             nodes=[NodeRecord.encode(node) for node in flowGraph.nodes]
#         )
#
#     def decode(self) -> FlowGraphData:
#         data = {
#             "edges": self.edges,
#             "nodes": [node.decode() for node in self.nodes],
#         }
#         return FlowGraphData.construct(_fields_set=self.__fields_set__, **data)
#
#
# class FlowRecord(NamedEntity):
#     description: Optional[str] = None
#     flowGraph: Optional[FlowGraphDataRecord] = None
#     enabled: Optional[bool] = True
#     projects: Optional[List[str]] = ["General"]
#
#     # Persistence
#
#     def storage(self) -> StorageCrud:
#         return StorageCrud("flow", FlowRecord, entity=self)
#
#     @staticmethod
#     def encode(flow: Flow) -> 'FlowRecord':
#         return FlowRecord(
#             id=flow.id,
#             description=flow.description,
#             name=flow.name,
#             enabled=flow.enabled,
#             flowGraph=FlowGraphDataRecord.encode(flow.flowGraph),
#             projects=flow.projects
#
#         )
#
#     def decode(self) -> Flow:
#         data = {
#             "id": self.id,
#             "name": self.name,
#             "description": self.description,
#             "enabled": self.enabled,
#             "projects": self.projects,
#             "flowGraph": self.flowGraph.decode() if self.flowGraph else None,
#         }
#         return Flow.construct(_fields_set=self.__fields_set__, **data)



from pprint import pprint

from tracardi_dot_notation.dot_accessor import DotAccessor

from tracardi.domain.context import Context
from tracardi.domain.event import Event
from tracardi.domain.flow import Flow
from tracardi.domain.profile import Profile
from tracardi.domain.session import Session
from tracardi.domain.source import Source
from tracardi.process_engine.tql.parser import Parser
from tracardi.process_engine.tql.transformer.expr_transformer import ExprTransformer
from tracardi.process_engine.tql.utils.dictonary import flatten

if __name__ == "__main__":

    data = {
        "n": 1,
        "a": {
            "b": 1,
            "c": [1,2,3],
            "d": {"aa": 1},
            "e": "test",
            'f': 1,
            'g': True,
            'h': None,
            'i': "2021-01-10"
        }
    }

    p = Parser(Parser.read('grammar/uql_expr.lark'), start='expr')
    # t = p.parse("a.b=1 and (a.c == 2 or a.c == [1,2,3])")
    # t = p.parse("datetime(a.i) between datetime(\"2020-01-01\") and datetime(\"2022-01-01\")")
    # t = p.parse("a.d.aa between 2 and 1")
    # t = p.parse("a.e == \"test\"")
    # t = p.parse("a.b == a.f")
    # t = p.parse("a.g == TRUE")
    # t = p.parse("a.h == null")
    # t = p.parse("profile@id == \"1\"")
    # t = p.parse("payload@a.h exists")
    # t = p.parse("payload@a.h == 1")
    t = p.parse("payload@n == 1")
    # pprint(t)

    profile = Profile(id="1")
    session = Session(id="2")
    payload = data
    source = Source(id="3", type="event")
    context = Context()
    event = Event(id="event-id", type="type", source=source, context=context, profile=profile, session=session)
    flow = Flow(id="flow-id", name="flow")
    dot = DotAccessor(profile, session, payload, event, flow)

    query = ExprTransformer(dot=dot).transform(t)
    pprint(query)
    # print(data)


from tracardi_plugin_sdk.domain.register import Plugin, Spec, MetaData
from tracardi_plugin_sdk.action_runner import ActionRunner
from tracardi_plugin_sdk.domain.result import Result

from tracardi.domain.profile_stats import ProfileStats


class IncreaseViewsAction(ActionRunner):

    def __init__(self, *args, **kwargs):
        pass

    async def run(self, payload):
        if self.profile.stats is None:
            self.profile.stats = ProfileStats()

        self.profile.increase_views()

        return Result(port="payload", value=payload)


def register() -> Plugin:
    return Plugin(
        start=False,
        spec=Spec(
            module='tracardi.process_engine.action.v1.increase_views_action',
            className='IncreaseViewsAction',
            inputs=["payload"],
            outputs=['payload'],
            init=None,
            manual="increase_views_action",
            version='0.1',
            license="MIT",
            author="Risto Kowaczewski"
        ),
        metadata=MetaData(
            name='Increase views',
            desc='Increases view field in profile and returns profile.',
            type='flowNode',
            width=200,
            height=100,
            icon='plus',
            group=["Stats"]
        )
    )



from datetime import datetime

import tzlocal

from tracardi.event_server.utils.timezone import get_timezone

dt = datetime.now()
print(dt.astimezone())
print(tzlocal.get_localzone().utcoffset(dt))
print()
dt = datetime.utcnow()
print(dt)
print(dt.astimezone())
print()
import pytz
my_date = datetime.now(pytz.timezone('Europe/Warsaw'))
print(my_date)
print(get_timezone())


from tracardi_plugin_sdk.domain.register import Plugin, Spec, MetaData
from tracardi_plugin_sdk.domain.result import Result
from tracardi_plugin_sdk.action_runner import ActionRunner
from deepdiff import DeepDiff
from tracardi.domain.event import Event
from tracardi.domain.profile import Profile
from tracardi.domain.session import Session
from tracardi_dot_notation.dot_accessor import DotAccessor

from tracardi.process_engine.tql.utils.dictonary import flatten


class CopyTraitAction(ActionRunner):

    def __init__(self, **kwargs):
        if 'copy' not in kwargs:
            raise ValueError("Please define copy in config section.")
        if not isinstance(kwargs['copy'], dict):
            raise ValueError("Please define copy as dictionary not {}.".format(type(kwargs['copy'])))

        self.mapping = kwargs['copy']

    async def run(self, payload: dict):

        dot = DotAccessor(
            self.profile,
            self.session,
            payload if isinstance(payload, dict) else None,
            self.event,
            self.flow)

        for destination, value in self.mapping.items():
            dot[destination] = value

        if not isinstance(dot.profile['traits']['private'], dict):
            raise ValueError(
                "Error when setting profile@traits.private to value `{}`. Private must have key:value pair. "
                "E.g. `name`: `{}`".format(dot.profile['traits']['private'], dot.profile['traits']['private']))

        if not isinstance(dot.profile['traits']['public'], dict):
            raise ValueError("Error when setting profile@traits.public to value `{}`. Public must have key:value pair. "
                             "E.g. `name`: `{}`".format(dot.profile['traits']['public'],
                                                        dot.profile['traits']['public']))

        profile = Profile(**dot.profile)
        event = Event(**dot.event)
        session = Session(**dot.session)

        flat_profile = flatten(profile.dict())
        flat_dot_profile = flatten(Profile(**dot.profile).dict())
        diff_result = DeepDiff(flat_dot_profile, flat_profile, exclude_paths=["root['metadata.time.insert']"])

        if diff_result and 'dictionary_item_removed' in diff_result:
            errors = [item.replace("root[", "profile[") for item in diff_result['dictionary_item_removed']]
            error_msg = "Some values were not added to profile. Profile schema seems not to have path: {}. " \
                        "This node is probably misconfigured.".format(errors)
            raise ValueError(error_msg)

        self.profile.replace(profile)
        self.session.replace(session)
        self.event.replace(event)

        return Result(port="payload", value=payload)


def register() -> Plugin:
    return Plugin(
        start=False,
        spec=Spec(
            module='tracardi.process_engine.action.v1.traits.copy_trait_action',
            className='CopyTraitAction',
            inputs=['payload'],
            outputs=["payload"],
            init={
                "copy": {
                    "target1": "source1",
                    "target2": "source2",
                }
            },
            version='0.1',
            license="MIT",
            author="Risto Kowaczewski"
        ),
        metadata=MetaData(
            name='Copy/Set Trait',
            desc='Returns payload with copied/set traits.',
            type='flowNode',
            width=100,
            height=100,
            icon='copy',
            group=["Traits"]
        )
    )



from pydantic import BaseModel
from tracardi.domain.value_object.save_result import SaveResult


class CollectResult(BaseModel):
    session: SaveResult
    events: SaveResult
    profile: SaveResult



from lark import Transformer


class Namespace:
    def __init__(self, namespace, transformer):
        self.namespace = namespace
        self.transformer = transformer
        self._namespace_size = len(self.namespace) if self.namespace else 0

    def __repr__(self):
        return "Namespace:`{}`, Transformer: `{}`".format(self.namespace, type(self.transformer).__name__)

    def get_method(self, item):
        if self.namespace and item[:self._namespace_size] == self.namespace:
            method = item[self._namespace_size:]
            try:
                method = getattr(self.transformer, method)
            except AttributeError as e:
                raise AttributeError("Method get_method returned {}".format(str(e)))
            if callable(method):
                return method
        return None


class TransformerNamespace(Transformer):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._namespaces = []

    def namespace(self, namespace, transformer):
        if not isinstance(namespace, str):
            raise ValueError("Namespace parameter is not string")

        self._namespaces.append(Namespace(namespace, transformer))

    def __getattr__(self, item):
        if self._namespaces:
            for n in self._namespaces:
                method = n.get_method(item)
                if method:
                    return method

        raise AttributeError(
                "Method `{}` not available in any of defined namespaces. Available namespaces are {}".format(item,
                                                                                                         self._namespaces))


if __name__ == "__main__":
    class A(TransformerNamespace):
        def a(self):
            print("A.a")


    class B(TransformerNamespace):
        def a(self):
            print("B.a")

        def b(self):
            print("B.b")


    t = TransformerNamespace()

    t.namespace("__a_", A())
    t.namespace("__b_", B())

    t.__a_a()



from .mock_storage import profiles


class ProfileStorageMock:

    async def load(self, profile_id):
        for item in profiles:
            if profile_id in item:
                return item[profile_id]
        return None

    async def create(self, data):
        for item in data:
            profiles.append({item['id']: item})



from .common_transformer import CommonTransformer
from .transformer_namespace import TransformerNamespace


class FunctionTransformer(TransformerNamespace):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.namespace("uql_common__", CommonTransformer())

    def function(self, args):
        return 'FUNCTION', args

    def FUNCTION_NAME(self, args):
        return 'FUNCTION_NAME', args.value

    def params(self, args):
        return 'PARAMS', args

    def value(self, args):
        return args[0]



from typing import List

import elasticsearch
import logging
from tracardi.domain.value_object.bulk_insert_result import BulkInsertResult
from tracardi.exceptions.exception import StorageException
import tracardi.service.storage.elastic_storage as storage
from tracardi.domain.storage_result import StorageResult

_logger = logging.getLogger("PersistenceService")


class PersistenceService:

    def __init__(self, storage: storage.ElasticStorage):
        self.storage = storage

    async def load(self, id: str):
        try:
            return await self.storage.load(id)
        except elasticsearch.exceptions.ElasticsearchException as e:
            raise StorageException(str(e))

    async def load_by(self, field: str, value: str) -> StorageResult:
        try:
            return StorageResult(await self.storage.load_by(field, value))
        except elasticsearch.exceptions.ElasticsearchException as e:
            raise StorageException(str(e))

    async def load_by_values(self, field_value_pairs: List[tuple], limit=1000) -> StorageResult:
        try:
            return StorageResult(await self.storage.load_by_values(field_value_pairs, limit))
        except elasticsearch.exceptions.ElasticsearchException as e:
            raise StorageException(str(e))

    async def delete_by(self, field: str, value: str) -> dict:
        try:
            return await self.storage.delete_by(field, value)
        except elasticsearch.exceptions.ElasticsearchException as e:
            raise StorageException(str(e))

    async def load_all(self) -> StorageResult:
        try:
            query = {
                "size": 100,
                "query": {
                    "match_all": {}
                }
            }
            result = await self.storage.search(query)
            return StorageResult(result)
        except elasticsearch.exceptions.ElasticsearchException as e:
            raise StorageException(str(e))

    async def upsert(self, data) -> BulkInsertResult:
        try:

            if not isinstance(data, list):
                if 'id' in data:
                    data["_id"] = data['id']
                payload = [data]
            else:
                # Add id
                for d in data:
                    if 'id' in d:
                        d["_id"] = d['id']
                payload = data

            return await self.storage.create(payload)

        except elasticsearch.exceptions.ElasticsearchException as e:
            raise StorageException(str(e))

    async def delete(self, id: str) -> dict:
        try:
            return await self.storage.delete(id)
        except elasticsearch.exceptions.ElasticsearchException as e:
            raise StorageException(str(e))

    async def filter(self, query: dict) -> StorageResult:
        try:
            return StorageResult(await self.storage.search(query))
        except elasticsearch.exceptions.NotFoundError:
            _logger.warning("No result found for query {}".format(query))
            return StorageResult()
        except elasticsearch.exceptions.ElasticsearchException as e:
            raise StorageException(str(e))

    async def query(self, query):
        try:
            return await self.storage.search(query)
        except elasticsearch.exceptions.ElasticsearchException as e:
            raise StorageException(str(e))

    async def translate(self, sql):
        try:
            return await self.storage.translate(sql)
        except elasticsearch.exceptions.ElasticsearchException as e:
            raise StorageException(str(e))

    async def refresh(self, params=None, headers=None):
        try:
            return await self.storage.refresh(params, headers)
        except elasticsearch.exceptions.ElasticsearchException as e:
            raise StorageException(str(e))

    async def flush(self, params=None, headers=None):
        try:
            return await self.storage.flush(params, headers)
        except elasticsearch.exceptions.ElasticsearchException as e:
            raise StorageException(str(e))

    async def sql(self, sql):
        try:
            query = await self.storage.translate(sql)
            if '_source' in query and query['_source'] == False:
                query['_source'] = True
            return StorageResult(await self.storage.search(query))
        except elasticsearch.exceptions.ElasticsearchException as e:
            raise StorageException(str(e))



from tracardi_dot_notation.dot_accessor import DotAccessor

from ...domain.context import Context
from ...domain.event import Event
from ...domain.flow import Flow
from ...domain.profile import Profile
from ...domain.session import Session
from ...domain.source import Source
from ...process_engine.tql.parser import Parser
from ...process_engine.tql.transformer.expr_transformer import ExprTransformer
from ...process_engine.tql.utils.dictonary import flatten

payload = {
    "a": {
        "b": 1,
        "c": [1, 2, 3],
        "d": {"aa": 1},
        "e": "test",
        'f': 1,
        'g': True,
        'h': None,
        'i': "2021-01-10"
    }
}

profile = Profile(id="1")
session = Session(id="2")
source = Source(id="3", type="event")
context = Context()
event = Event(id="event-id", type="type", source=source, context=context, profile=profile, session=session)
flow = Flow(id="flow-id", name="flow")
dot = DotAccessor(profile, session, payload, event, flow)

parser = Parser(Parser.read('grammar/uql_expr.lark'), start='expr')


def test_tql_operations():
    tree = parser.parse("payload@a.d.aa between 1 and 2")
    assert ExprTransformer(dot=dot).transform(tree)

    tree = parser.parse("payload@a.e == \"test\"")
    assert ExprTransformer(dot=dot).transform(tree)

    tree = parser.parse("payload@a.b == payload@a.f")
    assert ExprTransformer(dot=dot).transform(tree)

    tree = parser.parse("payload@a.g == True")
    assert ExprTransformer(dot=dot).transform(tree)

    tree = parser.parse("payload@a.h == null")
    assert ExprTransformer(dot=dot).transform(tree)

    tree = parser.parse("payload@a.b == 1")
    assert ExprTransformer(dot=dot).transform(tree)

    tree = parser.parse("payload@a.b >= 1")
    assert ExprTransformer(dot=dot).transform(tree)

    tree = parser.parse("payload@a.b <= 1")
    assert ExprTransformer(dot=dot).transform(tree)

    tree = parser.parse("payload@a.b > 0")
    assert ExprTransformer(dot=dot).transform(tree)

    tree = parser.parse("payload@a.b < 2")
    assert ExprTransformer(dot=dot).transform(tree)

    # tree = parser.parse("datetime(payload@a.i) == datetime(\"2021-01-10\")")
    # assert ExprTransformer(dot=dot).transform(tree)

    # tree = parser.parse("datetime(payload@a.i) <= datetime(\"2021-01-10\")")
    # assert ExprTransformer(dot=dot).transform(tree)

    # tree = parser.parse("datetime(payload@a.i) >= datetime(\"2021-01-10\")")
    # assert ExprTransformer(dot=dot).transform(tree)

    # tree = parser.parse("datetime(payload@a.i) < datetime(\"2021-01-11\")")
    # assert ExprTransformer(dot=dot).transform(tree)

    # tree = parser.parse("datetime(payload@a.i) > datetime(\"2021-01-09\")")
    # assert ExprTransformer(dot=dot).transform(tree)

    # tree = parser.parse("datetime(payload@a.i) < datetime(\"2021-01-10 00:00:01\")")
    # assert ExprTransformer(dot=dot).transform(tree)

    tree = parser.parse("payload@a.h is null")
    assert ExprTransformer(dot=dot).transform(tree)

    # tree = parser.parse("datetime(payload@a.i) between datetime(\"2020-01-01\") and datetime(\"2022-01-01\")")
    # assert ExprTransformer(dot=dot).transform(tree)

    tree = parser.parse("payload@a.h exists")
    assert ExprTransformer(dot=dot).transform(tree)

    tree = parser.parse("payload@a.h.h not exists")
    assert ExprTransformer(dot=dot).transform(tree)


def test_tql_false_operations():
    tree = parser.parse("payload@a.d.aa between 3 and 4")
    assert not ExprTransformer(dot=dot).transform(tree)

    tree = parser.parse("payload@a.e != \"test\"")
    assert not ExprTransformer(dot=dot).transform(tree)

    tree = parser.parse("payload@a.b != payload@a.f")
    assert not ExprTransformer(dot=dot).transform(tree)

    tree = parser.parse("payload@a.g != True")
    assert not ExprTransformer(dot=dot).transform(tree)

    tree = parser.parse("payload@a.h != null")
    assert not ExprTransformer(dot=dot).transform(tree)

    tree = parser.parse("payload@a.b != 1")
    assert not ExprTransformer(dot=dot).transform(tree)

    tree = parser.parse("payload@a.b > 1")
    assert not ExprTransformer(dot=dot).transform(tree)

    tree = parser.parse("payload@a.b < 1")
    assert not ExprTransformer(dot=dot).transform(tree)

    tree = parser.parse("payload@a.b > 0")
    assert ExprTransformer(dot=dot).transform(tree)

    tree = parser.parse("payload@a.b > 2")
    assert not ExprTransformer(dot=dot).transform(tree)

    # tree = parser.parse("datetime(payload@a.i) != datetime(\"2021-01-10\")")
    # assert not ExprTransformer(dot=dot).transform(tree)
    #
    # tree = parser.parse("datetime(a.i) < datetime(\"2021-01-10\")")
    # assert not ExprTransformer(dot=dot).transform(tree)
    #
    # tree = parser.parse("datetime(a.i) > datetime(\"2021-01-10\")")
    # assert not ExprTransformer(dot=dot).transform(tree)

    # tree = parser.parse("datetime(a.i) > datetime(\"2021-01-11\")")
    # assert not ExprTransformer(dot=dot).transform(tree)
    #
    # tree = parser.parse("datetime(a.i) < datetime(\"2021-01-09\")")
    # assert not ExprTransformer(dot=dot).transform(tree)
    #
    # tree = parser.parse("datetime(a.i) > datetime(\"2021-01-10 00:00:01\")")
    # assert not ExprTransformer(dot=dot).transform(tree)

    tree = parser.parse("payload@a.h != null")
    assert not ExprTransformer(dot=dot).transform(tree)

    # tree = parser.parse("datetime(a.i) between datetime(\"2022-01-01\") and datetime(\"2023-01-01\")")
    # assert not ExprTransformer(dot=dot).transform(tree)

    # tree = parser.parse("payload@a.h not exists")
    # assert not ExprTransformer(dot=dot).transform(tree)

    # tree = parser.parse("payload@a.h.h exists")
    # assert not ExprTransformer(dot=dot).transform(tree)


def test_tql_bool_operations():
    tree = parser.parse("payload@a.d.aa between 1 and 2 and payload@a.e == \"test\"")
    assert ExprTransformer(dot=dot).transform(tree)

    tree = parser.parse("payload@a.d.aa between 1 and 2 or payload@a.e != \"test\"")
    assert ExprTransformer(dot=dot).transform(tree)

    tree = parser.parse("payload@a.d.aa between 2 and 3 and payload@a.e != \"test\"")
    assert not ExprTransformer(dot=dot).transform(tree)




from pydantic import BaseModel


class BulkInsertResult(BaseModel):
    saved: int = 0
    errors: list = []
    ids: list = []

    def __add__(self, other):
        if not isinstance(other, BulkInsertResult):
            raise ValueError("BulkInsertResult can add only other BulkInsertResult")

        self.saved += other.saved
        self.errors += other.errors
        self.ids += other.ids

        return self

    def is_nothing_saved(self) -> bool:
        return len(self.ids) == 0



import asyncio
from uuid import uuid4

from tracardi.domain.consent import Consent
from tracardi.domain.profile import Profile
from tracardi.domain.session import Session
from tracardi.domain.source import Source


async def main(obj):
    obj = obj(id=str(uuid4()))
    result = await obj.storage().load()
    assert result is None
    await obj.storage().save()
    consent_loaded = await obj.storage().load()
    assert consent_loaded.id == obj.id
    await obj.storage().delete()
    result = await obj.storage().load()
    assert result is None

loop = asyncio.get_event_loop()
for o in [
    Consent,
    Profile,
    Session,
    # Flow,
    # Segment,
    Source,
    # UserConsent
]:
    loop.run_until_complete(main(o))
loop.close()



from tracardi.service.singleton import Singleton


def test_singleton():
    class Single(metaclass=Singleton):

        def __init__(self):
            self.prop = 1

    x = Single()
    y = Single()

    x.prop = 2
    assert y.prop == 2
    assert y == x



from typing import Optional

from tracardi.domain.entity import Entity
from tracardi.domain.third_party_processor import ThirdPartyProcessor
from tracardi.domain.time_range_query import DatetimePayload
from tracardi.service.storage.crud import StorageCrud


class Consent(Entity):
    description: Optional[str] = None
    third_party_processor: Optional[ThirdPartyProcessor] = None
    revoke: Optional[DatetimePayload] = None

    def storage(self) -> StorageCrud:
        return StorageCrud("consent", Consent, entity=self)
